from __future__ import annotations

import re
from typing import TYPE_CHECKING, Generator
if TYPE_CHECKING:
    from ..account import Account

import traceback
import json
import logging
from bs4 import BeautifulSoup

from ..common import exceptions
from .events import *


logger = logging.getLogger("FunPayAPI.runner")


class Runner:
    """
    Класс для получения новых событий FunPay.

    :param account: экземпляр аккаунта (должен быть инициализирован с помощью метода :meth:`FunPayAPI.account.Account.get`).
    :type account: :class:`FunPayAPI.account.Account`

    :param disable_message_requests: отключить ли запросы для получения истории чатов?\n
        Если `True`, :meth:`FunPayAPI.updater.runner.Runner.listen` не будет возвращать события
        :class:`FunPayAPI.updater.events.NewMessageEvent`.\n
        Из событий, связанных с чатами, будут возвращаться только:\n
        * :class:`FunPayAPI.updater.events.InitialChatEvent`\n
        * :class:`FunPayAPI.updater.events.ChatsListChangedEvent`\n
        * :class:`FunPayAPI.updater.events.LastChatMessageChangedEvent`\n
    :type disable_message_requests: :obj:`bool`, опционально

    :param disabled_order_requests: отключить ли запросы для получения списка заказов?\n
        Если `True`, :meth:`FunPayAPI.updater.runner.Runner.listen` не будет возвращать события
        :class:`FunPayAPI.updater.events.InitialOrderEvent`, :class:`FunPayAPI.updater.events.NewOrderEvent`,
        :class:`FunPayAPI.updater.events.OrderStatusChangedEvent`.\n
        Из событий, связанных с заказами, будет возвращаться только
        :class:`FunPayAPI.updater.events.OrdersListChangedEvent`.
    :type disabled_order_requests: :obj:`bool`, опционально
    """
    def __init__(self, account: Account, disable_message_requests: bool = False,
                 disabled_order_requests: bool = False):
        # todo добавить события и исключение событий о новых покупках (не продажах!)
        if not account.is_initiated:
            raise exceptions.AccountNotInitiatedError()
        if account.runner:
            raise Exception("К аккаунту уже привязан Runner!")  # todo

        self.make_msg_requests: bool = False if disable_message_requests else True
        """Делать ли доп. запросы для получения всех новых сообщений изменившихся чатов?"""
        self.make_order_requests: bool = False if disabled_order_requests else True
        """Делать ли доп запросы для получения все новых / изменившихся заказов?"""

        self.__first_request = True
        self.__last_msg_event_tag = utils.random_tag()
        self.__last_order_event_tag = utils.random_tag()

        self.saved_orders: dict[str, types.OrderShortcut] = {}
        """Сохраненные состояния заказов ({ID заказа: экземпляр types.OrderShortcut})."""

        self.last_messages: dict[int, list[str, str | None]] = {}
        """ID последний сообщений ({ID чата: (текст сообщения (до 250 символов), время сообщения)})."""

        self.init_messages: dict[int, str] = {}
        """Текста инит. чатов (для generate_new_message_events)."""

        self.by_bot_ids: dict[int, list[int]] = {}
        """ID сообщений, отправленных с помощью self.account.send_message ({ID чата: [ID сообщения, ...]})."""

        self.last_messages_ids: dict[int, int] = {}
        """ID последних сообщений в чатах ({ID чата: ID последнего сообщения})."""

        self.account: Account = account
        """Экземпляр аккаунта, к которому привязан Runner."""
        self.account.runner = self

        self.__msg_time_re = re.compile(r"\d{2}:\d{2}")

    def get_updates(self) -> dict:
        """
        Запрашивает список событий FunPay.

        :return: ответ FunPay.
        :rtype: :obj:`dict`
        """
        orders = {
            "type": "orders_counters",
            "id": self.account.id,
            "tag": self.__last_order_event_tag,
            "data": False
        }
        chats = {
            "type": "chat_bookmarks",
            "id": self.account.id,
            "tag": self.__last_msg_event_tag,
            "data": False
        }
        payload = {
            "objects": json.dumps([orders, chats]),
            "request": False,
            "csrf_token": self.account.csrf_token
        }
        headers = {
            "accept": "*/*",
            "content-type": "application/x-www-form-urlencoded; charset=UTF-8",
            "x-requested-with": "XMLHttpRequest"
        }

        response = self.account.method("post", "runner/", headers, payload, raise_not_200=True)
        json_response = response.json()
        logger.debug(f"Получены данные о событиях: {json_response}")
        return json_response

    def parse_updates(self, updates: dict) -> list[InitialChatEvent | ChatsListChangedEvent |
                                                   LastChatMessageChangedEvent | NewMessageEvent | InitialOrderEvent |
                                                   OrdersListChangedEvent | NewOrderEvent | OrderStatusChangedEvent]:
        """
        Парсит ответ FunPay и создает события.

        :param updates: результат выполнения :meth:`FunPayAPI.updater.runner.Runner.get_updates`
        :type updates: :obj:`dict`

        :return: список событий.
        :rtype: :obj:`list` of :class:`FunPayAPI.updater.events.InitialChatEvent`,
            :class:`FunPayAPI.updater.events.ChatsListChangedEvent`,
            :class:`FunPayAPI.updater.events.LastChatMessageChangedEvent`,
            :class:`FunPayAPI.updater.events.NewMessageEvent`, :class:`FunPayAPI.updater.events.InitialOrderEvent`,
            :class:`FunPayAPI.updater.events.OrdersListChangedEvent`,
            :class:`FunPayAPI.updater.events.NewOrderEvent`,
            :class:`FunPayAPI.updater.events.OrderStatusChangedEvent`
        """
        events = []
        for obj in updates["objects"]:
            if obj.get("type") == "chat_bookmarks":
                events.extend(self.parse_chat_updates(obj))
            elif obj.get("type") == "orders_counters":
                events.extend(self.parse_order_updates(obj))

        if self.__first_request:
            self.__first_request = False
        return events

    def parse_chat_updates(self, obj) -> list[InitialChatEvent | ChatsListChangedEvent | LastChatMessageChangedEvent |
                                              NewMessageEvent]:
        """
        Парсит события, связанные с чатами.

        :param obj: словарь из результата выполнения :meth:`FunPayAPI.updater.runner.Runner.get_updates`, где
            "type" == "chat_bookmarks".
        :type obj: :obj:`dict`

        :return: список событий, связанных с чатами.
        :rtype: :obj:list of :class:`FunPayAPI.updater.events.InitialChatEvent`,
            :class:`FunPayAPI.updater.events.ChatsListChangedEvent`,
            :class:`FunPayAPI.updater.events.LastChatMessageChangedEvent`,
            :class:`FunPayAPI.updater.events.NewMessageEvent`
        """
        events, lcmc_events = [], []
        self.__last_msg_event_tag = obj.get("tag")
        parser = BeautifulSoup(obj["data"]["html"], "html.parser")
        chats = parser.find_all("a", {"class": "contact-item"})

        # Получаем все изменившиеся чаты
        for chat in chats:
            chat_id = int(chat["data-id"])
            # Если чат удален админами - скип.
            if not (last_msg_text := chat.find("div", {"class": "contact-item-message"})):
                continue

            last_msg_text = last_msg_text.text
            if last_msg_text.startswith(self.account.bot_character):
                last_msg_text = last_msg_text.replace(self.account.bot_character, "", 1)
            last_msg_time = chat.find("div", {"class": "contact-item-time"}).text

            # Если текст последнего сообщения совпадает с сохраненным
            if chat_id in self.last_messages and self.last_messages[chat_id][0] == last_msg_text:
                # Если есть сохраненное время сообщения для данного чата
                if self.last_messages[chat_id][1]:
                    # Если время ласт сообщения не имеет формат ЧЧ:ММ или совпадает с сохраненным - скип чата
                    if not self.__msg_time_re.fullmatch(last_msg_time) or self.last_messages[chat_id][1] == last_msg_time:
                        continue
                # Если нет сохраненного времени сообщения для данного чата - скип чата
                else:
                    continue

            unread = True if "unread" in chat.get("class") else False
            chat_with = chat.find("div", {"class": "media-user-name"}).text
            chat_obj = types.ChatShortcut(chat_id, chat_with, last_msg_text, unread, str(chat))
            self.account.add_chats([chat_obj])
            self.last_messages[chat_id] = [last_msg_text, last_msg_time]

            if self.__first_request:
                events.append(InitialChatEvent(self.__last_msg_event_tag, chat_obj))
                self.init_messages[chat_id] = last_msg_text
                continue

            lcmc_events.append(LastChatMessageChangedEvent(self.__last_msg_event_tag, chat_obj))

        if lcmc_events:
            events.append(ChatsListChangedEvent(self.__last_msg_event_tag))

        if not self.make_msg_requests:
            events.extend(lcmc_events)
            return events

        while lcmc_events:
            chats_pack = lcmc_events[:10]
            del lcmc_events[:10]
            chats_data = {i.chat.id: i.chat.name for i in chats_pack}
            new_msg_events = self.generate_new_message_events(chats_data)
            for i in chats_pack:
                events.append(i)
                if new_msg_events.get(i.chat.id):
                    events.extend(new_msg_events[i.chat.id])
        return events

    def generate_new_message_events(self, chats_data: dict[int, str]) -> dict[int, list[NewMessageEvent]]:
        
        """
        Получает историю переданных чатов и генерирует события новых сообщений.

        :param chats_data: ID чатов и никнеймы собеседников (None, если никнейм неизвестен)
            Например: {48392847: "SLLMK", 58392098: "Amongus", 38948728: None}
        :type chats_data: :obj:`dict` {:obj:`int`: :obj:`str` or :obj:`None`}

        :return: словарь с событиями новых сообщений в формате {ID чата: [список событий]}
        :rtype: :obj:`dict` {:obj:`int`: :obj:`list` of :class:`FunPayAPI.updater.events.NewMessageEvent`}
        """
        attempts = 3
        while attempts:
            attempts -= 1
            try:
                chats = self.account.get_chats_histories(chats_data)
                break
            except exceptions.RequestFailedError as e:
                logger.error(e)
            except Exception as e:
                # Получаем информацию о стеке вызовов
                # tb = traceback.format_exc()
                logger.error(f"Не удалось получить истории чатов {list(chats_data.keys())}. Ошибка: {e}")
                logger.debug("TRACEBACK", exc_info=True)
            time.sleep(1)
        else:
            logger.error(f"Не удалось получить истории чатов {list(chats_data.keys())}: превышено кол-во попыток.")
            return {}

        result = {}

        for cid in chats:
            messages = chats[cid]
            result[cid] = []
            self.by_bot_ids[cid] = self.by_bot_ids.get(cid) or []

            # Удаляем все сообщения, у которых ID меньше сохраненного последнего сообщения
            if self.last_messages_ids.get(cid):
                messages = [i for i in messages if i.id > self.last_messages_ids[cid]]
            if not messages:
                continue

            # Отмечаем все сообщения, отправленные с помощью Account.send_message()
            if self.by_bot_ids.get(cid):
                for i in messages:
                    if not i.by_bot and i.id in self.by_bot_ids[cid]:
                        i.by_bot = True

            stack = MessageEventsStack()

            # Если нет сохраненного ID последнего сообщения
            if not self.last_messages_ids.get(cid):
                # Если данный чат был доступен при первом запросе и есть сохраненное последнее сообщение,
                # то ищем новые сообщения относительно последнего сохраненного текста
                if init_msg_text := self.init_messages.get(cid):
                    del self.init_messages[cid]
                    temp = []
                    for i in reversed(messages):
                        if i.text[:250] == init_msg_text:
                            break
                        temp.append(i)
                    messages = list(reversed(temp))

                # Если данного чата не было при первом запросе, в результат добавляем только ласт сообщение истории.
                else:
                    messages = messages[-1:]

            self.last_messages_ids[cid] = messages[-1].id  # Перезаписываем ID последнего сообщение
            self.by_bot_ids[cid] = [i for i in self.by_bot_ids[cid] if i > self.last_messages_ids[cid]]  # чистим память

            for msg in messages:
                event = NewMessageEvent(self.__last_msg_event_tag, msg, stack)
                stack.add_events([event])
                result[cid].append(event)
        return result

    def parse_order_updates(self, obj) -> list[InitialOrderEvent | OrdersListChangedEvent | NewOrderEvent |
                                               OrderStatusChangedEvent]:
        """
        Парсит события, связанные с продажами.

        :param obj: словарь из результата выполнения :meth:`FunPayAPI.updater.runner.Runner.get_updates`, где
            "type" == "orders_counters".
        :type obj: :obj:`dict`

        :return: список событий, связанных с продажами.
        :rtype: :obj:`list` of :class:`FunPayAPI.updater.events.InitialOrderEvent`,
            :class:`FunPayAPI.updater.events.OrdersListChangedEvent`,
            :class:`FunPayAPI.updater.events.NewOrderEvent`,
            :class:`FunPayAPI.updater.events.OrderStatusChangedEvent`
        """
        events = []
        self.__last_order_event_tag = obj.get("tag")
        if not self.__first_request:
            events.append(OrdersListChangedEvent(self.__last_order_event_tag,
                                                 obj["data"]["buyer"], obj["data"]["seller"]))
        if not self.make_order_requests:
            return events

        attempts = 3
        while attempts:
            attempts -= 1
            try:
                orders_list = self.account.get_sells()
                break
            except exceptions.RequestFailedError as e:
                logger.error(e)
            except:
                logger.error("Не удалось обновить список заказов.")
                logger.debug("TRACEBACK", exc_info=True)
            time.sleep(1)
        else:
            logger.error("Не удалось обновить список продаж: превышено кол-во попыток.")
            return events

        for order in orders_list[1]:
            if order.id not in self.saved_orders:
                if self.__first_request:
                    events.append(InitialOrderEvent(self.__last_order_event_tag, order))
                else:
                    events.append(NewOrderEvent(self.__last_order_event_tag, order))
                    if order.status == types.OrderStatuses.CLOSED:
                        events.append(OrderStatusChangedEvent(self.__last_order_event_tag, order))
                self.update_order(order)

            elif order.status != self.saved_orders[order.id].status:
                events.append(OrderStatusChangedEvent(self.__last_order_event_tag, order))
                self.update_order(order)
        return events

    def update_last_message(self, chat_id: int, message_text: str | None, message_time: str | None = None):
        """
        Обновляет сохраненный текст последнего сообщения чата.

        :param chat_id: ID чата.
        :type chat_id: :obj:`int`

        :param message_text: текст сообщения (если `None`, заменяется за "Изображение").
        :type message_text: :obj:`str` or :obj:`None`

        :param message_time: время отправки сообщения в формате ЧЧ:ММ. Используется исключительно Runner'ом.
        :type message_time: :obj:`str` or :obj:`None`, опционально
        """
        if message_text is None:
            message_text = "Изображение"
        self.last_messages[chat_id] = [message_text[:250], message_time]

    def update_order(self, order: types.OrderShortcut):
        """
        Обновляет сохраненное состояние переданного заказа.

        :param order: экземпляр заказа, который нужно обновить.
        :type order: :class:`FunPayAPI.types.OrderShortcut`
        """
        self.saved_orders[order.id] = order

    def mark_as_by_bot(self, chat_id: int, message_id: int):
        """
        Помечает сообщение с переданным ID, как отправленный с помощью :meth:`FunPayAPI.account.Account.send_message`.

        :param chat_id: ID чата.
        :type chat_id: :obj:`int`

        :param message_id: ID сообщения.
        :type message_id: :obj:`int`
        """
        if self.by_bot_ids.get(chat_id) is None:
            self.by_bot_ids[chat_id] = [message_id]
        else:
            self.by_bot_ids[chat_id].append(message_id)

    def listen(self, requests_delay: int | float = 6.0,
               ignore_exceptions: bool = True) -> Generator[InitialChatEvent | ChatsListChangedEvent |
                                                            LastChatMessageChangedEvent | NewMessageEvent |
                                                            InitialOrderEvent | OrdersListChangedEvent | NewOrderEvent |
                                                            OrderStatusChangedEvent]:
        """
        Бесконечно отправляет запросы для получения новых событий.

        :param requests_delay: задержка между запросами (в секундах).
        :type requests_delay: :obj:`int` or :obj:`float`, опционально

        :param ignore_exceptions: игнорировать ошибки?
        :type ignore_exceptions: :obj:`bool`, опционально

        :return: генератор событий FunPay.
        :rtype: :obj:`Generator` of :class:`FunPayAPI.updater.events.InitialChatEvent`,
            :class:`FunPayAPI.updater.events.ChatsListChangedEvent`,
            :class:`FunPayAPI.updater.events.LastChatMessageChangedEvent`,
            :class:`FunPayAPI.updater.events.NewMessageEvent`, :class:`FunPayAPI.updater.events.InitialOrderEvent`,
            :class:`FunPayAPI.updater.events.OrdersListChangedEvent`,
            :class:`FunPayAPI.updater.events.NewOrderEvent`,
            :class:`FunPayAPI.updater.events.OrderStatusChangedEvent`
        """
        while True:
            try:
                updates = self.get_updates()
                events = self.parse_updates(updates)
                for event in events:
                    yield event
            except Exception as e:
                if not ignore_exceptions:
                    raise e
                else:
                    logger.error("Произошла ошибка при получении событий. "
                                 "(ничего страшного, если это сообщение появляется нечасто).")
                    logger.debug("TRACEBACK", exc_info=True)
            time.sleep(requests_delay)



from __future__ import annotations
from enum import Enum


class EventTypes(Enum):
    """
    В данном классе перечислены все типы событий FunPayAPI.
    """
    INITIAL_CHAT = 0
    """Обнаружен чат (при первом запросе Runner'а)."""

    CHATS_LIST_CHANGED = 1
    """Список чатов и/или последнее сообщение одного/нескольких чатов изменилось."""

    LAST_CHAT_MESSAGE_CHANGED = 2
    """В чате изменилось последнее сообщение."""

    NEW_MESSAGE = 3
    """Обнаружено новое сообщение в истории чата."""

    INITIAL_ORDER = 4
    """Обнаружен заказ (при первом запросе Runner'а)."""

    ORDERS_LIST_CHANGED = 5
    """Список заказов и/или статус одного/нескольких заказов изменился."""

    NEW_ORDER = 6
    """Новый заказ."""

    ORDER_STATUS_CHANGED = 7
    """Статус заказа изменился."""


class MessageTypes(Enum):
    """
    В данном классе перечислены все типы сообщений.
    """
    NON_SYSTEM = 0
    """Несистемное сообщение."""

    ORDER_PURCHASED = 1
    """Покупатель X оплатил заказ #Y. Лот. X, не забудьте потом нажать кнопку «Подтвердить выполнение заказа»."""

    ORDER_CONFIRMED = 2
    """Покупатель X подтвердил успешное выполнение заказа #Y и отправил деньги продавцу Z."""

    NEW_FEEDBACK = 3
    """Покупатель X написал отзыв к заказу #Y."""

    FEEDBACK_CHANGED = 4
    """Покупатель X изменил отзыв к заказу #Y."""

    FEEDBACK_DELETED = 5
    """Покупатель X удалил отзыв к заказу #Y."""

    NEW_FEEDBACK_ANSWER = 6
    """Продавец Z ответил на отзыв к заказу #Y."""

    FEEDBACK_ANSWER_CHANGED = 7
    """Продавец Z изменил ответ на отзыв к заказу #Y."""

    FEEDBACK_ANSWER_DELETED = 8
    """Продавец Z удалил ответ на отзыв к заказу #Y."""

    ORDER_REOPENED = 9
    """Заказ #Y открыт повторно."""

    REFUND = 10
    """Продавец Z вернул деньги покупателю X по заказу #Y."""

    PARTIAL_REFUND = 11
    """Часть средств по заказу #Y возвращена покупателю."""

    ORDER_CONFIRMED_BY_ADMIN = 12
    """Администратор A подтвердил успешное выполнение заказа #Y и отправил деньги продавцу Z."""

    DISCORD = 13
    """Вы можете перейти в Discord. Внимание: общение за пределами сервера FunPay считается нарушением правил."""


class OrderStatuses(Enum):
    """
    В данном классе перечислены все состояния заказов.
    """
    PAID = 0
    """Заказ оплачен и ожидает выполнения."""
    CLOSED = 1
    """Заказ закрыт."""
    REFUNDED = 2
    """Средства по заказу возвращены."""


class SubCategoryTypes(Enum):
    """
    В данном классе перечислены все типы подкатегорий.
    """
    COMMON = 0
    """Подкатегория со стандартными лотами."""
    CURRENCY = 1
    """Подкатегория с лотами игровой валюты (их нельзя поднимать)."""


class Currency(Enum):
    """
    В данном классе перечислены все типы валют баланса FunPay.
    """
    USD = 0
    """Доллар"""
    RUB = 1
    """Рубль"""
    EUR = 2
    """Евро"""


class Wallet(Enum):
    """
    В данном классе перечислены все кошельки для вывода средств с баланса FunPay.
    """
    QIWI = 0
    """Qiwi кошелек."""
    BINANCE = 1
    """Binance Pay."""
    TRC = 2
    """USDT TRC20."""
    CARD_RUB = 3
    """Рублевая банковская карта."""
    CARD_USD = 4
    """Долларовая банковская карта."""
    CARD_EUR = 5
    """Евро банковская карта."""
    WEBMONEY = 6
    """WebMoney WMZ."""
    YOUMONEY = 7
    """ЮMoney."""



"""
В данном модуле описаны все типы пакета FunPayAPI
"""
from __future__ import annotations
from typing import Literal, overload, Optional
from .common.utils import RegularExpressions
from .common.enums import MessageTypes, OrderStatuses, SubCategoryTypes
import datetime


class ChatShortcut:
    """
    Данный класс представляет виджет чата со страницы https://funpay.com/chat/

    :param id_: ID чата.
    :type id_: :obj:`int`

    :param name: название чата (никнейм собеседника).
    :type name: :obj:`str`

    :param last_message_text: текст последнего сообщения в чате (макс. 250 символов).
    :type last_message_text: :obj:`str`

    :param unread: флаг "непрочитанности" (`True`, если чат не прочитан (оранжевый). `False`, если чат прочитан).
    :type unread: :obj:`bool`

    :param html: HTML код виджета чата.
    :type html: :obj:`str`

    :param determine_msg_type: определять ли тип последнего сообщения?
    :type determine_msg_type: :obj:`bool`, опционально
    """
    def __init__(self, id_: int, name: str, last_message_text: str,
                 unread: bool, html: str, determine_msg_type: bool = True):
        self.id: int = id_
        """ID чата."""
        self.name: str | None = name if name else None
        """Название чата (никнейм собеседника)."""
        self.last_message_text: str = last_message_text
        """Текст последнего сообщения в чате (макс. 250 символов)."""
        self.unread: bool = unread
        """Флаг \"непрочитанности\" (если True - в чате есть непрочитанные сообщения)."""
        self.last_message_type: MessageTypes | None = None if not determine_msg_type else self.get_last_message_type()
        """Тип последнего сообщения."""
        self.html: str = html
        """HTML код виджета чата."""

    def get_last_message_type(self) -> MessageTypes:
        """
        Определяет тип последнего сообщения в чате на основе регулярных выражений из MessageTypesRes.

        !Внимание! Результат определения типа сообщения данным методом не является правильным в 100% случаев, т.к. он
        основан на сравнении с регулярными выражениями.
        Возможны "ложные срабатывание", если пользователь напишет "поддельное" сообщение, которое совпадет с одним из
        регулярных выражений.

        :return: тип последнего сообщения.
        :rtype: :class:`FunPayAPI.common.enums.MessageTypes`
        """
        res = RegularExpressions()
        if self.last_message_text == res.DISCORD:
            return MessageTypes.DISCORD

        if res.ORDER_PURCHASED.findall(self.last_message_text) and res.ORDER_PURCHASED2.findall(self.last_message_text):
            return MessageTypes.ORDER_PURCHASED

        if res.ORDER_ID.search(self.last_message_text) is None:
            return MessageTypes.NON_SYSTEM

        # Регулярные выражения выставлены в порядке от самых часто-используемых к самым редко-используемым
        sys_msg_types = {
            MessageTypes.ORDER_CONFIRMED: res.ORDER_CONFIRMED,
            MessageTypes.NEW_FEEDBACK: res.NEW_FEEDBACK,
            MessageTypes.NEW_FEEDBACK_ANSWER: res.NEW_FEEDBACK_ANSWER,
            MessageTypes.FEEDBACK_CHANGED: res.FEEDBACK_CHANGED,
            MessageTypes.FEEDBACK_DELETED: res.FEEDBACK_DELETED,
            MessageTypes.REFUND: res.REFUND,
            MessageTypes.FEEDBACK_ANSWER_CHANGED: res.FEEDBACK_ANSWER_CHANGED,
            MessageTypes.FEEDBACK_ANSWER_DELETED: res.FEEDBACK_ANSWER_DELETED,
            MessageTypes.ORDER_CONFIRMED_BY_ADMIN: res.ORDER_CONFIRMED_BY_ADMIN,
            MessageTypes.PARTIAL_REFUND: res.PARTIAL_REFUND,
            MessageTypes.ORDER_REOPENED: res.ORDER_REOPENED
        }

        for i in sys_msg_types:
            if sys_msg_types[i].search(self.last_message_text):
                return i
        else:
            return MessageTypes.NON_SYSTEM

    def __str__(self):
        return self.last_message_text


class Chat:
    """
    Данный класс представляет личный чат.

    :param id_: ID чата.
    :type id_: :obj:`int`

    :param name: название чата (никнейм собеседника).
    :type name: :obj:`str`

    :param looking_link: ссылка на лот, который смотрит собеседник.
    :type looking_link: :obj:`str` or :obj:`None`

    :param looking_text: название лота, который смотрит собеседник.
    :type looking_text: :obj:`str` or :obj:`None`

    :param html: HTML код чата.
    :type html: :obj:`str`

    :param messages: последние 100 сообщений чата.
    :type messages: :obj:`list` of :class:`FunPayAPI.types.Message` or :obj:`None`
    """
    def __init__(self, id_: int, name: str, looking_link: str | None, looking_text: str | None,
                 html: str, messages: Optional[list[Message]] = None):
        self.id: int = id_
        """ID чата."""
        self.name: str = name
        """Название чата (никнейм собеседника)."""
        self.looking_link: str | None = looking_link
        """Ссылка на лот, который в данный момент смотрит собеседник."""
        self.looking_text: str | None = looking_text
        """Название лота, который в данный момент смотрит собеседник."""
        self.html: str = html
        """HTML код чата."""
        self.messages: list[Message] = messages or []
        """Последние 100 сообщений чата."""


class Message:
    """
    Данный класс представляет отдельное сообщение.

    :param id_: ID сообщения.
    :type id_: :obj:`int`

    :param text: текст сообщения (если есть).
    :type text: :obj:`str` or :obj:`None`

    :param chat_id: ID чата, в котором находится данное сообщение.
    :type chat_id: :obj:`int` or :obj:`str`

    :param chat_name: название чата, в котором находится данное сообщение.
    :type chat_name: :obj:`str` or :obj:`None`

    :param author: никнейм автора сообщения.
    :type author: :obj:`str`, or :obj:`None`

    :param author_id: ID автора сообщения.
    :type author_id: :obj:`int`

    :param html: HTML код сообщения.
    :type html: :obj:`str`

    :param image_link: ссылка на изображение из сообщения (если есть).
    :type image_link: :obj:`str` or :obj:`None`, опционально

    :param determine_msg_type: определять ли тип сообщения.
    :type determine_msg_type: :obj:`bool`, опционально
    """
    def __init__(self, id_: int, text: str | None, chat_id: int | str, chat_name: str | None,
                 author: str | None, author_id: int, html: str,
                 image_link: str | None = None, determine_msg_type: bool = True, badge_text: Optional[str] = None):
        self.id: int = id_
        """ID сообщения."""
        self.text: str | None = text
        """Текст сообщения."""
        self.chat_id: int | str = chat_id
        """ID чата."""
        self.chat_name: str | None = chat_name
        """Название чата."""
        self.type: MessageTypes | None = None if not determine_msg_type else self.get_message_type()
        """Тип сообщения."""
        self.author: str | None = author
        """Автор сообщения."""
        self.author_id: int = author_id
        """ID автора сообщения."""
        self.html: str = html
        """HTML-код сообщения."""
        self.image_link: str | None = image_link
        """Ссылка на изображение в сообщении (если оно есть)."""
        self.by_bot: bool = False
        """Отправлено ли сообщение с помощью :meth:`FunPayAPI.Account.send_message`?"""
        self.badge: str | None = badge_text
        """Текст бэйджика тех. поддержки."""

    def get_message_type(self) -> MessageTypes:
        """
        Определяет тип сообщения на основе регулярных выражений из MessageTypesRes.

        Внимание! Данный способ определения типа сообщения не является 100% правильным, т.к. он основан на сравнении с
        регулярными выражениями. Возможно ложное "срабатывание", если пользователь напишет "поддельное" сообщение,
        которое совпадет с одним из регулярных выражений.
        Рекомендуется делать проверку на author_id == 0.

        :return: тип последнего сообщения в чате.
        :rtype: :class:`FunPayAPI.common.enums.MessageTypes`
        """
        if not self.text:
            return MessageTypes.NON_SYSTEM

        res = RegularExpressions()
        if self.text == res.DISCORD:
            return MessageTypes.DISCORD

        if res.ORDER_PURCHASED.findall(self.text) and res.ORDER_PURCHASED2.findall(self.text):
            return MessageTypes.ORDER_PURCHASED

        if res.ORDER_ID.search(self.text) is None:
            return MessageTypes.NON_SYSTEM

        # Регулярные выражения выставлены в порядке от самых часто-используемых к самым редко-используемым
        sys_msg_types = {
            MessageTypes.ORDER_CONFIRMED: res.ORDER_CONFIRMED,
            MessageTypes.NEW_FEEDBACK: res.NEW_FEEDBACK,
            MessageTypes.NEW_FEEDBACK_ANSWER: res.NEW_FEEDBACK_ANSWER,
            MessageTypes.FEEDBACK_CHANGED: res.FEEDBACK_CHANGED,
            MessageTypes.FEEDBACK_DELETED: res.FEEDBACK_DELETED,
            MessageTypes.REFUND: res.REFUND,
            MessageTypes.FEEDBACK_ANSWER_CHANGED: res.FEEDBACK_ANSWER_CHANGED,
            MessageTypes.FEEDBACK_ANSWER_DELETED: res.FEEDBACK_ANSWER_DELETED,
            MessageTypes.ORDER_CONFIRMED_BY_ADMIN: res.ORDER_CONFIRMED_BY_ADMIN,
            MessageTypes.PARTIAL_REFUND: res.PARTIAL_REFUND,
            MessageTypes.ORDER_REOPENED: res.ORDER_REOPENED
        }

        for i in sys_msg_types:
            if sys_msg_types[i].search(self.text):
                return i
        else:
            return MessageTypes.NON_SYSTEM

    def __str__(self):
        return self.text if self.text is not None else self.image_link if self.image_link is not None else ""


class OrderShortcut:
    """
    Данный класс представляет виджет заказа со страницы https://funpay.com/orders/trade

    :param id_: ID заказа.
    :type id_: :obj:`str`

    :param description: описание заказа.
    :type description: :obj:`str`

    :param price: цена заказа.
    :type price: :obj:`float`

    :param buyer_username: никнейм покупателя.
    :type buyer_username: :obj:`str`

    :param buyer_id: ID покупателя.
    :type buyer_id: :obj:`int`

    :param status: статус заказа.
    :type status: :class:`FunPayAPI.common.enums.OrderStatuses`

    :param date: дата создания заказа.
    :type date: :class:`datetime.datetime`

    :param subcategory_name: название подкатегории, к которой относится заказ.
    :type subcategory_name: :obj:`str`

    :param html: HTML код виджета заказа.
    :type html: :obj:`str`

    :param dont_search_amount: не искать кол-во товара.
    :type dont_search_amount: :obj:`bool`, опционально
    """
    def __init__(self, id_: str, description: str, price: float,
                 buyer_username: str, buyer_id: int, status: OrderStatuses,
                 date: datetime.datetime, subcategory_name: str, html: str, dont_search_amount: bool = False):
        self.id: str = id_ if not id_.startswith("#") else id_[1:]
        """ID заказа."""
        self.description: str = description
        """Описание заказа."""
        self.price: float = price
        """Цена заказа."""
        self.amount: int | None = self.parse_amount() if not dont_search_amount else None
        """Кол-во товаров."""
        self.buyer_username: str = buyer_username
        """Никнейм покупателя."""
        self.buyer_id: int = buyer_id
        """ID покупателя."""
        self.status: OrderStatuses = status
        """Статус заказа."""
        self.date: datetime.datetime = date
        """Дата создания заказа."""
        self.subcategory_name: str = subcategory_name
        """Название подкатегории, к которой относится заказ."""
        self.html: str = html
        """HTML код виджета заказа."""

    def parse_amount(self) -> int:
        """
        Парсит кол-во купленного товара (ищет подстроку по регулярному выражению).

        :return: кол-во купленного товара.
        :rtype: :obj:`int`
        """
        res = RegularExpressions()
        result = res.PRODUCTS_AMOUNT.findall(self.description)
        if result:
            return int(result[0].split(" ")[0])
        return 1

    def __str__(self):
        return self.description


class Order:
    """
    Данный класс представляет заказ со страницы https://funpay.com/orders/<ORDER_ID>/

    :param id_: ID заказа.
    :type id_: :obj:`str`

    :param status: статус заказа.
    :type status: :class:`FunPayAPI.common.enums.OrderStatuses`

    :param subcategory: подкатегория, к которой относится заказ.
    :type subcategory: :class:`FunPayAPI.types.SubCategory`

    :param short_description: краткое описание (название) заказа.
    :type short_description: :obj:`str` or :obj:`None`

    :param full_description: полное описание заказа.
    :type full_description: :obj:`str` or :obj:`None`

    :param sum_: сумма заказа.
    :type sum_: :obj:`float`

    :param buyer_id: ID покупателя.
    :type buyer_id: :obj:`int`

    :param buyer_username: никнейм покупателя.
    :type buyer_username: :obj:`str`

    :param seller_id: ID продавца.
    :type seller_id: :obj:`int`

    :param seller_username: никнейм продавца.
    :type seller_username: :obj:`str`

    :param html: HTML код заказа.
    :type html: :obj:`str`

    :param review: объект отзыва на заказ.
    :type review: :class:`FunPayAPI.types.Review` or :obj:`None`
    """
    def __init__(self, id_: str, status: OrderStatuses, subcategory: SubCategory, short_description: str | None,
                 full_description: str | None, sum_: float,
                 buyer_id: int, buyer_username: str,
                 seller_id: int, seller_username: str,
                 html: str, review: Review | None):
        self.id: str = id_ if not id_.startswith("#") else id_[1:]
        """ID заказа."""
        self.status: OrderStatuses = status
        """Статус заказа."""
        self.subcategory: SubCategory = subcategory
        """Подкатегория, к которой относится заказ."""
        self.short_description: str | None = short_description
        """Краткое описание (название) заказа. То же самое, что и Order.title."""
        self.title: str | None = short_description
        """Краткое описание (название) заказа. То же самое, что и Order.short_description."""
        self.full_description: str | None = full_description
        """Полное описание заказа."""
        self.sum: float = sum_
        """Сумма заказа."""
        self.buyer_id: int = buyer_id
        """ID покупателя."""
        self.buyer_username: str = buyer_username
        """Никнейм покупателя."""
        self.seller_id: int = seller_id
        """ID продавца."""
        self.seller_username: str = seller_username
        """Никнейм продавца."""
        self.html: str = html
        """HTML код заказа."""
        self.review: Review | None = review
        """Объект отзыва заказа."""


class Category:
    """
    Класс, описывающий категорию (игру).

    :param id_: ID категории (game_id / data-id).
    :type id_: :obj:`int`

    :param name: название категории (игры).
    :type name: :obj:`str`

    :param subcategories: подкатегории.
    :type subcategories: :obj:`list` of :class:`FunPayAPI.types.SubCategory` or :obj:`None`, опционально
    """
    def __init__(self, id_: int, name: str, subcategories: list[SubCategory] | None = None):
        self.id: int = id_
        """ID категории (game_id / data-id)."""
        self.name: str = name
        """Название категории (игры)."""
        self.__subcategories: list[SubCategory] = subcategories or []
        """Список подкатегорий."""
        self.__sorted_subcategories: dict[SubCategoryTypes, dict[int, SubCategory]] = {
            SubCategoryTypes.COMMON: {},
            SubCategoryTypes.CURRENCY: {}
        }
        for i in self.__subcategories:
            self.__sorted_subcategories[i.type][i.id] = i

    def add_subcategory(self, subcategory: SubCategory):
        """
        Добавляет подкатегорию в список подкатегорий.

        :param subcategory: объект подкатегории.
        :type subcategory: :class:`FunPayAPI.types.SubCategory`
        """
        if subcategory not in self.__subcategories:
            self.__subcategories.append(subcategory)
            self.__sorted_subcategories[subcategory.type][subcategory.id] = subcategory

    def get_subcategory(self, subcategory_type: SubCategoryTypes, subcategory_id: int) -> SubCategory | None:
        """
        Возвращает объект подкатегории.

        :param subcategory_type: тип подкатегории.
        :type subcategory_type: :class:`FunPayAPI.common.enums.SubCategoryTypes`

        :param subcategory_id: ID подкатегории.
        :type subcategory_id: :obj:`int`

        :return: объект подкатегории или None, если подкатегория не найдена.
        :rtype: :class:`FunPayAPI.types.SubCategory` or :obj:`None`
        """
        return self.__sorted_subcategories[subcategory_type].get(subcategory_id)

    def get_subcategories(self) -> list[SubCategory]:
        """
        Возвращает все подкатегории данной категории (игры).

        :return: все подкатегории данной категории (игры).
        :rtype: :obj:`list` of :class:`FunPayAPI.types.SubCategory`
        """
        return self.__subcategories

    def get_sorted_subcategories(self) -> dict[SubCategoryTypes, dict[int, SubCategory]]:
        """
        Возвращает все подкатегории данной категории (игры) в виде словаря {type: {ID: подкатегория}}.

        :return: все подкатегории данной категории (игры) в виде словаря {type: ID: подкатегория}}.
        :rtype: :obj:`dict` {:class:`FunPayAPI.common.enums.SubCategoryTypes`: :obj:`dict` {:obj:`int`, :class:`FunPayAPI.types.SubCategory`}}
        """
        return self.__sorted_subcategories


class SubCategory:
    """
    Класс, описывающий подкатегорию.

    :param id_: ID подкатегории.
    :type id_: :obj:`int`

    :param name: название подкатегории.
    :type name: :obj:`str`

    :param type_: тип лотов подкатегории.
    :type type_: :class:`FunPayAPI.common.enums.SubCategoryTypes`

    :param category: родительская категория (игра).
    :type category: :class:`FunPayAPI.types.Category`
    """
    def __init__(self, id_: int, name: str, type_: SubCategoryTypes, category: Category):
        self.id: int = id_
        """ID подкатегории."""
        self.name: str = name
        """Название подкатегории."""
        self.type: SubCategoryTypes = type_
        """Тип подкатегории."""
        self.category: Category = category
        """Родительская категория (игра)."""
        self.fullname: str = f"{self.name} {self.category.name}"
        """Полное название подкатегории."""
        self.public_link: str = f"https://funpay.com/chips/{id_}/" if type_ is SubCategoryTypes.CURRENCY else \
            f"https://funpay.com/lots/{id_}/"
        """Публичная ссылка на список лотов подкатегории."""
        self.private_link: str = f"{self.public_link}trade"
        """Приватная ссылка на список лотов подкатегории (для редактирования лотов)."""


class LotFields:
    """
    Класс, описывающий поля лота со страницы редактирования лота.

    :param lot_id: ID лота.
    :type lot_id: :obj:`int`

    :param fields: словарь с полями.
    :type fields: :obj:`dict`
    """
    def __init__(self, lot_id: int, fields: dict):
        self.lot_id: int = lot_id
        """ID лота."""
        self.__fields: dict = fields
        """Поля лота."""

        self.title_ru: str = self.__fields.get("fields[summary][ru]")
        """Русское краткое описание (название) лота."""
        self.title_en: str = self.__fields.get("fields[summary][en]")
        """Английское краткое описание (название) лота."""
        self.description_ru: str = self.__fields.get("fields[desc][ru]")
        """Русское полное описание лота."""
        self.description_en: str = self.__fields.get("fields[desc][en]")
        """Английское полное описание лота."""
        self.amount: int | None = int(i) if (i := self.__fields.get("amount")) else None
        """Кол-во товара."""
        self.price: float = float(i) if (i := self.__fields.get("price")) else None
        """Цена за 1шт."""
        self.active: bool = "active" in self.__fields
        """Активен ли лот."""
        self.deactivate_after_sale: bool = "deactivate_after_sale[]" in self.__fields
        """Деактивировать ли лот после продажи."""

    @property
    def fields(self) -> dict[str, str]:
        """
        Возвращает все поля лота в виде словаря.

        :return: все поля лота в виде словаря.
        :rtype: :obj:`dict` {:obj:`str`: :obj:`str`}
        """
        return self.__fields

    def edit_fields(self, fields: dict[str, str]):
        """
        Редактирует переданные поля лота.

        :param fields: поля лота, которые нужно заменить, и их значения.
        :type fields: obj:`dict` {:obj:`str`: :obj:`str`}
        """
        self.__fields.update(fields)

    def set_fields(self, fields: dict):
        """
        Сбрасывает текущие поля лота и устанавливает переданные.
        !НЕ РЕДАКТИРУЕТ СВОЙСТВА ЭКЗЕМЛПЯРА!

        :param fields: поля лота.
        :type fields: :obj:`dict` {:obj:`str`: :obj:`str`}
        """
        self.__fields = fields

    def renew_fields(self) -> LotFields:
        """
        Обновляет :py:obj:`~__fields` (возвращается в методе :meth:`FunPayAPI.types.LotFields.get_fields`),
        основываясь на свойствах экземпляра.
        Необходимо вызвать перед сохранением лота на FunPay после изменения любого свойства экземпляра.

        :return: экземпляр класса :class:`FunPayAPI.types.LotFields` с новыми полями лота.
        :rtype: :class:`FunPayAPI.types.LotFields`
        """
        self.__fields["fields[summary][ru]"] = self.title_ru
        self.__fields["fields[summary][en]"] = self.title_en
        self.__fields["fields[desc][ru]"] = self.description_ru
        self.__fields["fields[desc][en]"] = self.description_en
        self.__fields["price"] = str(self.price) if self.price is not None else ""
        self.__fields["deactivate_after_sale"] = "on" if self.deactivate_after_sale else ""
        self.__fields["active"] = "on" if self.active else ""
        self.__fields["amount"] = self.amount if self.amount is not None else ""
        return self


class LotShortcut:
    """
    Данный класс представляет виджет лота.

    :param id_: ID лота.
    :type id_: :obj:`int` or :obj:`str`

    :param server: название сервера (если указан в лоте).
    :type server: :obj:`str` or :obj:`None`

    :param description: краткое описание (название) лота.
    :type description: :obj:`str` or :obj:`None`

    :param price: цена лота.
    :type price: :obj:`float`

    :param subcategory: подкатегория лота.
    :type subcategory: :class:`FunPayAPI.types.SubCategory`

    :param html: HTML код виджета лота.
    :type html: :obj:`str`
    """
    def __init__(self, id_: int | str, server: str | None,
                 description: str | None, price: float, subcategory: SubCategory, html: str):
        self.id: int | str = id_
        if isinstance(self.id, str) and self.id.isnumeric():
            self.id = int(self.id)
        """ID лота."""
        self.server: str | None = server
        """Название сервера (если указан)."""
        self.description: str | None = description
        """Краткое описание (название) лота."""
        self.title: str | None = description
        """Краткое описание (название) лота."""
        self.price: float = price
        """Цена лота."""
        self.subcategory: SubCategory = subcategory
        """Подкатегория лота."""
        self.html: str = html
        """HTML-код виджета лота."""
        self.public_link: str = f"https://funpay.com/chips/offer?id={self.id}" \
            if self.subcategory.type is SubCategoryTypes.CURRENCY else f"https://funpay.com/lots/offer?id={self.id}"
        """Публичная ссылка на лот."""


class UserProfile:
    """
    Данный класс представляет пользователя FunPay.

    :param id_: ID пользователя.
    :type id_: :obj:`int`

    :param username: никнейм пользователя.
    :type username: :obj:`str`

    :param profile_photo: ссылка на фото профиля.
    :type profile_photo: :obj:`str`

    :param online: онлайн ли пользователь?
    :type online: :obj:`bool`

    :param banned: заблокирован ли пользователь?
    :type banned: :obj:`bool`

    :param html: HTML код страницы пользователя.
    :type html: :obj:`str`
    """
    def __init__(self, id_: int, username: str, profile_photo: str, online: bool, banned: bool, html: str):
        self.id: int = id_
        """ID пользователя."""
        self.username: str = username
        """Никнейм пользователя."""
        self.profile_photo: str = profile_photo
        """Ссылка на фото профиля."""
        self.online: bool = online
        """Онлайн ли пользователь."""
        self.banned: bool = banned
        """Заблокирован ли пользователь."""
        self.html: str = html
        """HTML код страницы пользователя."""
        self.__lots: list[LotShortcut] = []
        """Все лоты пользователя."""
        self.__lots_ids: dict[int | str, LotShortcut] = {}
        """Все лоты пользователя в виде словаря {ID: лот}}"""
        self.__sorted_by_subcategory_lots: dict[SubCategory, dict[int | str, LotShortcut]] = {}
        """Все лоты пользователя в виде словаря {подкатегория: {ID: лот}}"""
        self.__sorted_by_subcategory_type_lots: dict[SubCategoryTypes, dict[int | str, LotShortcut]] = {
            SubCategoryTypes.COMMON: {},
            SubCategoryTypes.CURRENCY: {}
        }

    def get_lot(self, lot_id: int | str) -> LotShortcut | None:
        """
        Возвращает объект лота со страницы пользователя.

        :param lot_id: ID лота.
        :type lot_id: :obj:`int` or :obj:`str`

        :return: объект лота со страницы пользователя или `None`, если объект не найден.
        :rtype: :class:`FunPayAPI.types.LotShortcut` or :obj:`None`
        """
        if isinstance(lot_id, str) and lot_id.isnumeric():
            return self.__lots_ids.get(int(lot_id))
        return self.__lots_ids.get(lot_id)

    def get_lots(self) -> list[LotShortcut]:
        """
        Возвращает список всех лотов пользователя.

        :return: список всех лотов пользователя.
        :rtype: :obj:`list` of :class:`FunPayAPI.types.LotShortcut`
        """
        return self.__lots

    @overload
    def get_sorted_lots(self, mode: Literal[1]) -> dict[int | str, LotShortcut]:
        ...

    @overload
    def get_sorted_lots(self, mode: Literal[2]) -> dict[SubCategory, dict[int | str, LotShortcut]]:
        ...

    @overload
    def get_sorted_lots(self, mode: Literal[3]) -> dict[SubCategoryTypes, dict[int | str, LotShortcut]]:
        ...

    def get_sorted_lots(self, mode: Literal[1, 2, 3]) -> dict[int | str, LotShortcut] |\
                                                         dict[SubCategory, dict[int | str, LotShortcut]] |\
                                                         dict[SubCategoryTypes, dict[int | str, LotShortcut]]:
        """
        Возвращает список всех лотов пользователя в виде словаря.

        :param mode: вариант словаря.\n
            1 - {ID: лот}\n
            2 - {подкатегория: {ID: лот}}\n
            3 - {тип лота: {ID: лот}}

        :return: список всех лотов пользователя в виде словаря.
        :rtype: :obj:`dict` {:obj:`int` or :obj:`str`: :class:`FunPayAPI.types.LotShortcut`} (`mode==1`) \n
            :obj:`dict` {:class:`FunPayAPI.types.SubCategory`: :obj:`dict` {:obj:`int` or :obj:`str`: :class:`FunPayAPI.types.LotShortcut`}} (`mode==2`) \n
            :obj:`dict` {:class:`FunPayAPI.common.enums.SubCategoryTypes`: :obj:`dict` {:obj:`int` or :obj:`str`: :class:`FunPayAPI.types.LotShortcut`}} (`mode==3`)
        """
        if mode == 1:
            return self.__lots_ids
        elif mode == 2:
            return self.__sorted_by_subcategory_lots
        else:
            return self.__sorted_by_subcategory_type_lots

    def add_lot(self, lot: LotShortcut):
        """
        Добавляет лот в список лотов.

        :param lot: объект лота.
        """
        if lot in self.__lots:
            return

        self.__lots.append(lot)
        self.__lots_ids[lot.id] = lot
        if lot.subcategory not in self.__sorted_by_subcategory_lots:
            self.__sorted_by_subcategory_lots[lot.subcategory] = {}
        self.__sorted_by_subcategory_lots[lot.subcategory][lot.id] = lot
        self.__sorted_by_subcategory_type_lots[lot.subcategory.type][lot.id] = lot

    def get_common_lots(self) -> list[LotShortcut]:
        """
        Возвращает список стандартных лотов со страницы пользователя.

        :return: Список стандартных лотов со страницы пользователя.
        :rtype: :obj:`list` of :class:`FunPayAPI.types.LotShortcut`
        """
        return list(self.__sorted_by_subcategory_type_lots[SubCategoryTypes.COMMON].values())

    def get_currency_lots(self) -> list[LotShortcut]:
        """
        Возвращает список лотов-валют со страницы пользователя.

        :return: список лотов-валют со страницы пользователя.
        :rtype: :obj:`list` of :class:`FunPayAPI.types.LotShortcut`
        """
        return list(self.__sorted_by_subcategory_type_lots[SubCategoryTypes.CURRENCY].values())

    def __str__(self):
        return self.username


class Review:
    """
    Данный класс представляет отзыв на заказ.

    :param stars: кол-во звезд в отзыве.
    :type stars: :obj:`int` or :obj:`None`

    :param text: текст отзыва.
    :type text: :obj:`str` or :obj:`None`

    :param reply: текст ответа на отзыв.
    :type reply: :obj:`str` or :obj:`None`

    :param anonymous: анонимный ли отзыв?
    :type anonymous: :obj:`bool`

    :param html: HTML код отзыва.
    :type html: :obj:`str`

    :param order_id: ID заказа, к которому относится отзыв.
    :type order_id: :obj:`str` or :obj:`None`, опционально

    :param author: автор отзыва.
    :type author: :obj:`str` or :obj:`None`, опционально

    :param author_id: ID автора отзыва.
    :type author_id: :obj:`int` or :obj:`None`, опционально
    """
    def __init__(self, stars: int | None, text: str | None, reply: str | None, anonymous: bool, html: str,
                 order_id: str | None = None, author: str | None = None, author_id: int | None = None):
        self.stars: int | None = stars
        """Кол-во звезде в отзыве."""
        self.text: str | None = text
        """Текст отзыва."""
        self.reply: str | None = reply
        """Текст ответа на отзыв."""
        self.anonymous: bool = anonymous
        """Анонимный ли отзыв?"""
        self.html: str = html
        """HTML код отзыва."""
        self.order_id: str | None = order_id[1:] if order_id and order_id.startswith("#") else order_id
        """ID заказа, к которому относится отзыв."""
        self.author: str | None = author
        """Автор отзыва."""
        self.author_id: int | None = author_id
        """ID автора отзыва."""


class Balance:
    """
    Данный класс представляет информацию о балансе аккаунта.

    :param total_rub: общий рублёвый баланс.
    :type total_rub: :obj:`float`

    :param available_rub: доступный к выводу рублёвый баланс.
    :type available_rub: :obj:`float`

    :param total_usd: общий долларовый баланс.
    :type total_usd: :obj:`float`

    :param available_usd: доступный к выводу долларовый баланс.
    :type available_usd: :obj:`float`

    :param total_eur: общий евро баланс.
    :param available_eur: :obj:`float`
    """
    def __init__(self, total_rub: float, available_rub: float, total_usd: float, available_usd: float,
                 total_eur: float, available_eur: float):
        self.total_rub: float = total_rub
        """Общий рублёвый баланс."""
        self.available_rub: float = available_rub
        """Доступный к выводу рублёвый баланс."""
        self.total_usd: float = total_usd
        """Общий долларовый баланс."""
        self.available_usd: float = available_usd
        """Доступный к выводу долларовый баланс."""
        self.total_eur: float = total_eur
        """Общий евро баланс."""
        self.available_eur: float = available_eur
        """Доступный к выводу евро баланс."""



"""
В данном модуле написаны вспомогательные функции.
"""

import string
import random
import re


MONTHS = {
    "января": 1,
    "февраля": 2,
    "марта": 3,
    "апреля": 4,
    "мая": 5,
    "июня": 6,
    "июля": 7,
    "августа": 8,
    "сентября": 9,
    "октября": 10,
    "ноября": 11,
    "декабря": 12
}


def random_tag() -> str:
    """
    Генерирует случайный тег для запроса (для runner'а).

    :return: сгенерированный тег.
    """
    return "".join(random.choice(string.digits + string.ascii_lowercase) for _ in range(10))


def parse_wait_time(response: str) -> int:
    """
    Парсит ответ FunPay на запрос о поднятии лотов.

    :param response: текст ответа.

    :return: Примерное время ожидание до следующего поднятия лотов (в секундах).
    """
    if response == "Подождите секунду.":
        return 2
    elif response == "Подождите минуту.":
        return 60
    elif response == "Подождите час.":
        return 3600
    elif "сек" in response:
        response = response.split()
        return int(response[1])
    elif "мин" in response:
        response = response.split()
        # ["Подождите", "n", "минут."]
        return (int(response[1])-1) * 60
    elif "час" in response:
        response = response.split()
        return (int(response[1])) * 3600
    else:
        return 10


class RegularExpressions(object):
    """
    В данном классе хранятся скомпилированные регулярные выражения, описывающие системные сообщения FunPay и прочие
    элементы текстов.
    Класс является singleton'ом.
    """
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "instance"):
            setattr(cls, "instance", super(RegularExpressions, cls).__new__(cls))
        return getattr(cls, "instance")

    def __init__(self):
        self.ORDER_PURCHASED = re.compile(r"Покупатель [a-zA-Z0-9]+ оплатил заказ #[A-Z0-9]{8}\.")
        """
        Скомпилированное регулярное выражение, описывающее сообщение об оплате заказа.
        Лучше всего использовать вместе с MessageTypesRes.ORDER_PURCHASED2
        """

        self.ORDER_PURCHASED2 = re.compile(r"[a-zA-Z0-9]+, не забудьте потом нажать кнопку "
                                           r"«Подтвердить выполнение заказа»\.")
        """
        Скомпилированное регулярное выражение, описывающее сообщение об оплате заказа (2).
        Лучше всего использовать вместе с MessageTypesRes.ORDER_PURCHASED
        """

        self.ORDER_CONFIRMED = re.compile(r"Покупатель [a-zA-Z0-9]+ подтвердил успешное выполнение "
                                          r"заказа #[A-Z0-9]{8} и отправил деньги продавцу [a-zA-Z0-9]+\.")
        """
        Скомпилированное регулярное выражение, описывающее сообщение о подтверждении выполнения заказа.
        """

        self.NEW_FEEDBACK = re.compile(r"Покупатель [a-zA-Z0-9]+ написал отзыв к заказу #[A-Z0-9]{8}\.")
        """
        Скомпилированное регулярное выражение, описывающее сообщение о новом отзыве.
        """

        self.FEEDBACK_CHANGED = re.compile(r"Покупатель [a-zA-Z0-9]+ изменил отзыв к заказу #[A-Z0-9]{8}\.")
        """
        Скомпилированное регулярное выражение, описывающее сообщение об изменении отзыва.
        """

        self.FEEDBACK_DELETED = re.compile(r"Покупатель [a-zA-Z0-9]+ удалил отзыв к заказу #[A-Z0-9]{8}\.")
        """
        Скомпилированное регулярное выражение, описывающее сообщение об удалении отзыва.
        """

        self.NEW_FEEDBACK_ANSWER = re.compile(r"Продавец [a-zA-Z0-9]+ ответил на отзыв к заказу #[A-Z0-9]{8}\.")
        """
        Скомпилированное регулярное выражение, описывающее сообщение о новом ответе на отзыв.
        """

        self.FEEDBACK_ANSWER_CHANGED = re.compile(r"Продавец [a-zA-Z0-9]+ изменил ответ на отзыв к "
                                                  r"заказу #[A-Z0-9]{8}\.")
        """
        Скомпилированное регулярное выражение, описывающее сообщение об изменении ответа на отзыв.
        """

        self.FEEDBACK_ANSWER_DELETED = re.compile(r"Продавец [a-zA-Z0-9]+ удалил ответ на отзыв к заказу "
                                                  r"#[A-Z0-9]{8}\.")
        """
        Скомпилированное регулярное выражение, описывающее сообщение об удалении ответа на отзыв.
        """

        self.ORDER_REOPENED = re.compile(r"Заказ #[A-Z0-9]{8} открыт повторно\.")
        """
        Скомпилированное регулярное выражение, описывающее сообщение о повтором открытии заказа.
        """

        self.REFUND = re.compile(r"Продавец [a-zA-Z0-9]+ вернул деньги покупателю [a-zA-Z0-9]+ "
                                 r"по заказу #[A-Z0-9]{8}\.")
        """
        Скомпилированное регулярное выражение, описывающее сообщение о возврате денежных средств.
        """

        self.PARTIAL_REFUND = re.compile(r"Часть средств по заказу #[A-Z0-9]{8} возвращена покупателю\.")
        """
        Скомпилированное регулярное выражение, описывающее сообщение частичном о возврате денежных средств.
        """

        self.ORDER_CONFIRMED_BY_ADMIN = re.compile(r"Администратор [a-zA-Z0-9]+ подтвердил успешное выполнение "
                                                   r"заказа #[A-Z0-9]{8} и отправил деньги продавцу [a-zA-Z0-9]+\.")
        """
        Скомпилированное регулярное выражение, описывающее сообщение о подтверждении выполнения заказа администратором.
        """

        self.ORDER_ID = re.compile(r"#[A-Z0-9]{8}")
        """
        Скомпилированное регулярное выражение, описывающее ID заказа.
        """

        self.ORDER_DATE = re.compile(r"\d{1,2} [а-я]+, \d{1,2}:\d{1,2}")
        """
        Скомпилированное регулярное выражение, описывающее дату заказа в формате <ДД месяца, ЧЧ:ММ>.
        """

        self.FULL_ORDER_DATE = re.compile(r"\d{1,2} [а-я]+ \d{4}, \d{1,2}:\d{1,2}")
        """
        Скомпилированное регулярное выражение, описывающее дату заказа в формате <ДД месяца ГГГГ, ЧЧ:ММ>.
        """

        self.DISCORD = "Вы можете перейти в Discord. " \
                       "Внимание: общение за пределами сервера FunPay считается нарушением правил."
        """
        Точный текст сообщения о предложении перехода в Discord.
        """

        self.PRODUCTS_AMOUNT = re.compile(r"\d+ шт\.")
        """
        Скомпилированное регулярное выражение, описывающее запись кол-ва товаров в заказе.
        """



"""
В данном модуле описаны все кастомные исключения, используемые в пакете FunPayAPI.
"""
import requests
from .. import types


class AccountNotInitiatedError(Exception):
    """
    Исключение, которое возбуждается, если предпринята попытка вызвать метод класса :class:`FunPayAPI.account.Account`
    без предварительного получения данных аккаунта с помощью метода :meth:`FunPayAPI.account.Account.get`.
    """
    def __init__(self):
        pass

    def __str__(self):
        return "Необходимо получить данные об аккаунте с помощью метода Account.get()"


class RequestFailedError(Exception):
    """
    Исключение, которое возбуждается, если статус код ответа != 200.
    """
    def __init__(self, response: requests.Response):
        """
        :param response: объект ответа.
        """
        self.response = response
        self.status_code = response.status_code
        self.url = response.request.url
        self.request_headers = response.request.headers
        if "cookie" in self.request_headers:
            self.request_headers["cookie"] = "HIDDEN"
        self.request_body = response.request.body
        self.log_response = False

    def short_str(self):
        return f"Ошибка запроса к {self.url}. (Статус-код: {self.status_code})"

    def __str__(self):
        msg = f"Ошибка запроса к {self.url} .\n" \
              f"Метод: {self.response.request.method} .\n" \
              f"Статус-код ответа: {self.status_code} .\n" \
              f"Заголовки запроса: {self.request_headers} .\n" \
              f"Тело запроса: {self.request_body} ."
        if self.log_response:
            msg += f"\n{self.response.content.decode()}"
        return msg


class UnauthorizedError(RequestFailedError):
    """
    Исключение, которое возбуждается, если не удалось найти идентифицирующий аккаунт элемент и / или произошло другое
    событие, указывающее на отсутствие авторизации.
    """
    def __init__(self, response):
        super(UnauthorizedError, self).__init__(response)

    def short_str(self):
        return "Не авторизирован (возможно, введен неверный golden_key?)."


class WithdrawError(RequestFailedError):
    """
    Исключение, которое возбуждается, если произошла ошибка при попытке вывести средства с аккаунта.
    """
    def __init__(self, response, error_message: str | None):
        super(WithdrawError, self).__init__(response)
        self.error_message = error_message
        if not self.error_message:
            self.log_response = True

    def short_str(self):
        return f"Произошла ошибка при выводе средств с аккаунта{f': {self.error_message}' if self.error_message else '.'}"


class RaiseError(RequestFailedError):
    """
    Исключение, которое возбуждается, если произошла ошибка при попытке поднять лоты.
    """
    def __init__(self, response, category: types.Category, error_message: str | None, wait_time: int | None):
        super(RaiseError, self).__init__(response)
        self.category = category
        self.error_message = error_message
        self.wait_time = wait_time

    def short_str(self):
        return f"Не удалось поднять лоты категории \"{self.category.name}\"" \
               f"{f': {self.error_message}' if self.error_message else '.'}"


class ImageUploadError(RequestFailedError):
    """
    Исключение, которое возбуждается, если произошла ошибка при выгрузке изображения.
    """
    def __init__(self, response: requests.Response, error_message: str | None):
        super(ImageUploadError, self).__init__(response)
        self.error_message = error_message
        if not self.error_message:
            self.log_response = True

    def short_str(self):
        return f"Произошла ошибка при выгрузке изображения{f': {self.error_message}' if self.error_message else '.'}"


class MessageNotDeliveredError(RequestFailedError):
    """
    Исключение, которое возбуждается, если при отправке сообщения произошла ошибка.
    """
    def __init__(self, response: requests.Response, error_message: str | None, chat_id: int):
        super(MessageNotDeliveredError, self).__init__(response)
        self.error_message = error_message
        self.chat_id = chat_id
        if not self.error_message:
            self.log_response = True

    def short_str(self):
        return f"Не удалось отправить сообщение в чат {self.chat_id}" \
               f"{f': {self.error_message}' if self.error_message else '.'}"


class FeedbackEditingError(RequestFailedError):
    """
    Исключение, которое возбуждается, если при добавлении / редактировании / удалении отзыва / ответа на отзыв
    произошла ошибка.
    """
    def __init__(self, response: requests.Response, error_message: str | None, order_id: str):
        super(FeedbackEditingError, self).__init__(response)
        self.error_message = error_message
        self.order_id = order_id
        if not self.error_message:
            self.log_response = True

    def short_str(self):
        return f"Не удалось изменить состояние отзыва / ответа на отзыв на заказ {self.order_id}" \
               f"{f': {self.error_message}' if self.error_message else '.'}"


class LotSavingError(RequestFailedError):
    """
    Исключение, которое возбуждается, если при сохранении лота произошла ошибка.
    """
    def __init__(self, response: requests.Response, error_message: str | None, lot_id: int):
        super(LotSavingError, self).__init__(response)
        self.error_message = error_message
        self.lot_id = lot_id
        if not self.error_message:
            self.log_response = True

    def short_str(self):
        return f"Не удалось сохранить лот {self.lot_id}" \
               f"{f': {self.error_message}' if self.error_message else '.'}"


class RefundError(RequestFailedError):
    """
    Исключение, которое возбуждается, если при возврате средств за заказ произошла ошибка.
    """
    def __init__(self, response: requests.Response, error_message: str | None, order_id: str):
        super(RefundError, self).__init__(response)
        self.error_message = error_message
        self.order_id = order_id
        if not self.error_message:
            self.log_response = True

    def short_str(self):
        return f"Не удалось вернуть средства по заказу {self.order_id}" \
               f"{f': {self.error_message}' if self.error_message else '.'}"



from __future__ import annotations
from typing import TYPE_CHECKING, Literal, Any, Optional, IO
if TYPE_CHECKING:
    from .updater.runner import Runner

from requests_toolbelt import MultipartEncoder
from bs4 import BeautifulSoup
from datetime import datetime, timedelta
import requests
import logging
import random
import string
import json
import time
import re

from . import types
from .common import exceptions, utils, enums


logger = logging.getLogger("FunPayAPI.account")
PRIVATE_CHAT_ID_RE = re.compile(r"users-\d+-\d+$")


class Account:
    """
    Класс для управления аккаунтом FunPay.

    :param golden_key: токен (golden_key) аккаунта.
    :type golden_key: :obj:`str`

    :param user_agent: user-agent браузера, с которого был произведен вход в аккаунт.
    :type user_agent: :obj:`str`

    :param requests_timeout: тайм-аут ожидания ответа на запросы.
    :type requests_timeout: :obj:`int` or :obj:`float`

    :param proxy: прокси для запросов.
    :type proxy: :obj:`dict` {:obj:`str`: :obj:`str` or :obj:`None`
    """
    def __init__(self, golden_key: str, user_agent: str | None = None,
                 requests_timeout: int | float = 10, proxy: Optional[dict] = None):
        self.golden_key: str = golden_key
        """Токен (golden_key) аккаунта."""
        self.user_agent: str | None = user_agent
        """User-agent браузера, с которого был произведен вход в аккаунт."""
        self.requests_timeout: int | float = requests_timeout
        """Тайм-аут ожидания ответа на запросы."""
        self.proxy = proxy

        self.html: str | None = None
        """HTML основной страницы FunPay."""
        self.app_data: dict | None = None
        """Appdata."""
        self.id: int | None = None
        """ID аккаунта."""
        self.username: str | None = None
        """Никнейм аккаунта."""
        self.active_sales: int | None = None
        """Активные продажи."""
        self.active_purchases: int | None = None
        """Активные покупки."""

        self.csrf_token: str | None = None
        """CSRF токен."""
        self.phpsessid: str | None = None
        """PHPSESSID сессии."""
        self.last_update: int | None = None
        """Последнее время обновления аккаунта."""

        self.__initiated: bool = False

        self.__saved_chats: dict[int, types.ChatShortcut] = {}
        self.runner: Runner | None = None
        """Объект Runner'а."""

        self.__categories: list[types.Category] = []
        self.__sorted_categories: dict[int, types.Category] = {}

        self.__subcategories: list[types.SubCategory] = []
        self.__sorted_subcategories: dict[types.SubCategoryTypes, dict[int, types.SubCategory]] = {
            types.SubCategoryTypes.COMMON: {},
            types.SubCategoryTypes.CURRENCY: {}
        }

        self.__bot_character = "⁤"
        """Если сообщение начинается с этого символа, значит оно отправлено ботом."""

    def method(self, request_method: Literal["post", "get"], api_method: str, headers: dict, payload: Any,
               exclude_phpsessid: bool = False, raise_not_200: bool = False) -> requests.Response:
        """
        Отправляет запрос к FunPay. Добавляет в заголовки запроса user_agent и куки.

        :param request_method: метод запроса ("get" / "post").
        :type request_method: :obj:`str` `post` or `get`

        :param api_method: метод API / полная ссылка.
        :type api_method: :obj:`str`

        :param headers: заголовки запроса.
        :type headers: :obj:`dict`

        :param payload: полезная нагрузка.
        :type payload: :obj:`dict`

        :param exclude_phpsessid: исключить ли PHPSESSID из добавляемых куки?
        :type exclude_phpsessid: :obj:`bool`

        :param raise_not_200: возбуждать ли исключение, если статус код ответа != 200?
        :type raise_not_200: :obj:`bool`

        :return: объект ответа.
        :rtype: :class:`requests.Response`
        """
        headers["cookie"] = f"golden_key={self.golden_key}"
        headers["cookie"] += f"; PHPSESSID={self.phpsessid}" if self.phpsessid and not exclude_phpsessid else ""
        if self.user_agent:
            headers["user-agent"] = self.user_agent
        link = api_method if api_method.startswith("https://funpay.com") else "https://funpay.com/" + api_method
        response = getattr(requests, request_method)(link, headers=headers, data=payload, timeout=self.requests_timeout,
                                                     proxies=self.proxy or {})

        if response.status_code == 403:
            raise exceptions.UnauthorizedError(response)
        elif response.status_code != 200 and raise_not_200:
            raise exceptions.RequestFailedError(response)
        return response

    def get(self, update_phpsessid: bool = False) -> Account:
        """
        Получает / обновляет данные об аккаунте. Необходимо вызывать каждые 40-60 минут, дабы обновить
        :py:obj:`.Account.phpsessid`.

        :param update_phpsessid: обновить :py:obj:`.Account.phpsessid` или использовать старый.
        :type update_phpsessid: :obj:`bool`, опционально

        :return: объект аккаунта с обновленными данными.
        :rtype: :class:`FunPayAPI.account.Account`
        """
        response = self.method("get", "https://funpay.com", {}, {}, update_phpsessid, raise_not_200=True)

        html_response = response.content.decode()
        parser = BeautifulSoup(html_response, "html.parser")

        username = parser.find("div", {"class": "user-link-name"})
        if not username:
            raise exceptions.UnauthorizedError(response)

        self.username = username.text
        self.app_data = json.loads(parser.find("body").get("data-app-data"))
        self.id = self.app_data["userId"]
        self.csrf_token = self.app_data["csrf-token"]

        active_sales = parser.find("span", {"class": "badge badge-trade"})
        self.active_sales = int(active_sales.text) if active_sales else 0

        active_purchases = parser.find("span", {"class": "badge badge-orders"})
        self.active_purchases = int(active_purchases.text) if active_purchases else 0

        cookies = response.cookies.get_dict()
        if update_phpsessid or not self.phpsessid:
            self.phpsessid = cookies["PHPSESSID"]
        if not self.is_initiated:
            self.__setup_categories(html_response)

        self.last_update = int(time.time())
        self.html = html_response
        self.__initiated = True
        return self

    def get_subcategory_public_lots(self, subcategory_type: enums.SubCategoryTypes, subcategory_id: int) -> list[types.LotShortcut]:
        """
        Получает список всех опубликованных лотов переданной подкатегории.

        :param subcategory_type: тип подкатегории.
        :type subcategory_type: :class:`FunPayAPI.enums.SubCategoryTypes`

        :param subcategory_id: ID подкатегории.
        :type subcategory_id: :obj:`int`

        :return: список всех опубликованных лотов переданной подкатегории.
        :rtype: :obj:`list` of :class:`FunPayAPI.types.LotShortcut`
        """
        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()

        meth = f"lots/{subcategory_id}/" if subcategory_type is enums.SubCategoryTypes.COMMON else f"chips/{subcategory_id}/"
        response = self.method("get", meth, {"accept": "*/*"}, {}, raise_not_200=True)
        html_response = response.content.decode()
        parser = BeautifulSoup(html_response, "html.parser")

        username = parser.find("div", {"class": "user-link-name"})
        if not username:
            raise exceptions.UnauthorizedError(response)

        offers = parser.find_all("a", {"class": "tc-item"})
        if not offers:
            return []

        subcategory_obj = self.get_subcategory(subcategory_type, subcategory_id)
        result = []
        for offer in offers:
            offer_id = offer["href"].split("id=")[1]
            description = offer.find("div", {"class": "tc-desc-text"})
            description = description.text if description else None
            server = offer.find("div", {"class": "tc-server hidden-xxs"})
            if not server:
                server = offer.find("div", {"class": "tc-server hidden-xs"})
            server = server.text if server else None

            if subcategory_type is types.SubCategoryTypes.COMMON:
                price = float(offer.find("div", {"class": "tc-price"})["data-s"])
            else:
                price = float(offer.find("div", {"class": "tc-price"}).find("div").text.split()[0])
            lot_obj = types.LotShortcut(offer_id, server, description, price, subcategory_obj, str(offer))
            result.append(lot_obj)
        return result

    def get_all_user_lots(self, user_id: int) -> list[dict]:
        """
        Получает список всех опубликованных лотов переданной подкатегории.

        :param subcategory_type: тип подкатегории.
        :type subcategory_type: :class:`FunPayAPI.enums.SubCategoryTypes`

        :param subcategory_id: ID подкатегории.
        :type subcategory_id: :obj:`int`

        :return: список всех опубликованных лотов переданной подкатегории.
        :rtype: :obj:`list` of :class:`FunPayAPI.types.LotShortcut`
        """
        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()

        meth = f"users/{user_id}/"
        response = self.method("get", meth, {"accept": "*/*"}, {}, raise_not_200=True)
        html_response = response.content.decode()
        parser = BeautifulSoup(html_response, "html.parser")

        username = parser.find("div", {"class": "user-link-name"})
        if not username:
            raise exceptions.UnauthorizedError(response)

        offers = parser.find_all("a", {"class": "tc-item"})
        if not offers:
            return []


        result = []
        for offer in offers:
            offer_id = offer["href"].split("id=")[1]
            description = offer.find("div", {"class": "tc-desc-text"})
            description = description.text if description else None


            price = float('{:.1f}'.format(float(offer.find("div", {"class": "tc-price"})["data-s"])))

            price_unit = offer.find("span", {"class": "unit"})
            price_unit = price_unit.text if price_unit else None

            lot_obj = {"offer_id" : offer_id, "description" : description, "price" : price, "price_unit" : price_unit}
            result.append(lot_obj)
        return result

    def get_balance(self, lot_id: int = 28436563) -> types.Balance:
        """
        Получает информацию о балансе пользователя.

        :param lot_id: ID лота, на котором проверять баланс.
        :type lot_id: :obj:`int`, опционально

        :return: информацию о балансе пользователя.
        :rtype: :class:`FunPayAPI.types.Balance`
        """
        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()
        response = self.method("get", f"lots/offer?id={lot_id}", {"accept": "*/*"}, {}, raise_not_200=True)
        html_response = response.content.decode()
        parser = BeautifulSoup(html_response, "html.parser")

        username = parser.find("div", {"class": "user-link-name"})
        if not username:
            raise exceptions.UnauthorizedError(response)

        balances = parser.find("select", {"name": "method"})
        balance = types.Balance(float(balances["data-balance-total-rub"]), float(balances["data-balance-rub"]),
                                float(balances["data-balance-total-usd"]), float(balances["data-balance-usd"]),
                                float(balances["data-balance-total-eur"]), float(balances["data-balance-eur"]))
        return balance

    def get_chat_history(self, chat_id: int | str, last_message_id: int = 99999999999999999999999,
                         interlocutor_username: Optional[str] = None, from_id: int = 0) -> list[types.Message]:
        """
        Получает историю указанного чата (до 100 последних сообщений).

        :param chat_id: ID чата (или его текстовое обозначение).
        :type chat_id: :obj:`int` or :obj:`str`

        :param last_message_id: ID сообщения, с которого начинать историю (фильтр FunPay).
        :type last_message_id: :obj:`int`

        :param interlocutor_username: никнейм собеседника. Не нужно указывать для получения истории публичного чата.
            Так же не обязательно, но желательно указывать для получения истории личного чата.
        :type interlocutor_username: :obj:`str` or :obj:`None`, опционально.

        :param from_id: все сообщения с ID < переданного не попадут в возвращаемый список сообщений.
        :type from_id: :obj:`int`, опционально.

        :return: история указанного чата.
        :rtype: :obj:`list` of :class:`FunPayAPI.types.Message`
        """
        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()

        headers = {
            "accept": "*/*",
            "x-requested-with": "XMLHttpRequest"
        }
        payload = {
            "node": chat_id,
            "last_message": last_message_id
        }
        response = self.method("get", f"chat/history?node={chat_id}&last_message={last_message_id}",
                               headers, payload, raise_not_200=True)

        json_response = response.json()
        if not json_response.get("chat") or not json_response["chat"].get("messages"):
            return []
        if isinstance(chat_id, int):
            interlocutor_id = int(json_response["chat"]["node"]["name"].split("-")[2])
        else:
            interlocutor_id = None
        return self.__parse_messages(json_response["chat"]["messages"], chat_id, interlocutor_id,
                                     interlocutor_username, from_id)

    def get_chats_histories(self, chats_data: dict[int | str, str | None]) -> dict[int, list[types.Message]]:
        """
        Получает историю сообщений сразу нескольких чатов
        (до 50 сообщений на личный чат, до 25 сообщений на публичный чат).

        :param chats_data: ID чатов и никнеймы собеседников (None, если никнейм неизвестен)\n
            Например: {48392847: "SLLMK", 58392098: "Amongus", 38948728: None}
        :type chats_data: :obj:`dict` {:obj:`int` or :obj:`str`: :obj:`str` or :obj:`None`}

        :return: словарь с историями чатов в формате {ID чата: [список сообщений]}
        :rtype: :obj:`dict` {:obj:`int`: :obj:`list` of :class:`FunPayAPI.types.Message`}
        """
        headers = {
            "accept": "*/*",
            "content-type": "application/x-www-form-urlencoded; charset=UTF-8",
            "x-requested-with": "XMLHttpRequest"
        }
        objects = [{"type": "chat_node", "id": i, "tag": "00000000",
                    "data": {"node": i, "last_message": -1, "content": ""}} for i in chats_data]
        payload = {
            "objects": json.dumps(objects),
            "request": False,
            "csrf_token": self.csrf_token
        }
        response = self.method("post", "runner/", headers, payload, raise_not_200=True)
        json_response = response.json()

        result = {}
        for i in json_response["objects"]:
            if not i.get("data"):
                result[i.get("id")] = []
                continue
            if isinstance(i.get("id"), int):
                interlocutor_id = int(i["data"]["node"]["name"].split("-")[2])
                interlocutor_name = chats_data[i.get("id")]
            else:
                interlocutor_id = None
                interlocutor_name = None
            messages = self.__parse_messages(i["data"]["messages"], i.get("id"), interlocutor_id, interlocutor_name)
            result[i.get("id")] = messages
        return result

    def get_buyer_looking_at(self, chat_id: int | str) -> dict:

        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()

        headers = {
            "accept": "*/*",
            "x-requested-with": "XMLHttpRequest"
        }
        payload = {
            "node": chat_id,
        }
        response = self.method("get", f"chat/?node={chat_id}", headers, payload, raise_not_200=True)

        soup = BeautifulSoup(response.content, 'html.parser')
        param_item = soup.find('div', class_='param-item chat-panel')

        link = param_item.find('div').find('a') if param_item else None


        url = link.get('href') if link else None
        text = link.text.strip() if link else None

        return {"url": url, "text": text}


    def upload_image(self, image: str | IO[bytes]) -> int:
        """
        Выгружает изображение на сервер FunPay для дальнейшей отправки в качестве сообщения.
        Для отправки изображения в чат рекомендуется использовать метод :meth:`FunPayAPI.account.Account.send_image`.

        :param image: путь до изображения или представление изображения в виде байтов.
        :type image: :obj:`str` or :obj:`bytes`

        :return: ID изображения на серверах FunPay.
        :rtype: :obj:`int`
        """
        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()

        if isinstance(image, str):
            with open(image, "rb") as f:
                img = f.read()
        else:
            img = image

        fields = {
            'file': ("funpay_cardinal_image.png", img, "image/png"),
            'file_id': "0"
        }
        boundary = '----WebKitFormBoundary' + ''.join(random.sample(string.ascii_letters + string.digits, 16))
        m = MultipartEncoder(fields=fields, boundary=boundary)

        headers = {
            "accept": "*/*",
            "x-requested-with": "XMLHttpRequest",
            "content-type": m.content_type,
        }

        response = self.method("post", "file/addChatImage", headers, m)

        if response.status_code == 400:
            try:
                json_response = response.json()
                message = json_response.get("msg")
                raise exceptions.ImageUploadError(response, message)
            except requests.exceptions.JSONDecodeError:
                raise exceptions.ImageUploadError(response, None)
        elif response.status_code != 200:
            raise exceptions.RequestFailedError(response)

        if not (document_id := response.json().get("fileId")):
            raise exceptions.ImageUploadError(response, None)
        return int(document_id)

    def send_message(self, chat_id: int | str, text: Optional[str] = None, chat_name: Optional[str] = None,
                     image_id: Optional[int] = None, add_to_ignore_list: bool = True,
                     update_last_saved_message: bool = False) -> types.Message:
        """
        Отправляет сообщение в чат.

        :param chat_id: ID чата.
        :type chat_id: :obj:`int` or :obj:`str`

        :param text: текст сообщения.
        :type text: :obj:`str` or :obj:`None`, опционально

        :param chat_name: название чата (для возвращаемого объекта сообщения) (не нужно для отправки сообщения в публичный чат).
        :type chat_name: :obj:`str` or :obj:`None`, опционально

        :param image_id: ID изображения. Доступно только для личных чатов.
        :type image_id: :obj:`int` or :obj:`None`, опционально

        :param add_to_ignore_list: добавлять ли ID отправленного сообщения в игнорируемый список Runner'а?
        :type add_to_ignore_list: :obj:`bool`, опционально

        :param update_last_saved_message: обновлять ли последнее сохраненное сообщение на отправленное в Runner'е?
        :type update_last_saved_message: :obj:`bool`, опционально.

        :return: экземпляр отправленного сообщения.
        :rtype: :class:`FunPayAPI.types.Message`
        """
        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()

        headers = {
            "accept": "*/*",
            "content-type": "application/x-www-form-urlencoded; charset=UTF-8",
            "x-requested-with": "XMLHttpRequest"
        }
        request = {
            "action": "chat_message",
            "data": {"node": chat_id, "last_message": -1, "content": text}
        }

        if image_id is not None:
            request["data"]["image_id"] = image_id
            request["data"]["content"] = ""
        else:
            request["data"]["content"] = f"{self.__bot_character}{text}" if text else ""

        objects = [
            {
                "type": "chat_node",
                "id": chat_id,
                "tag": "00000000",
                "data": {"node": chat_id, "last_message": -1, "content": ""}
            }
        ]
        payload = {
            "objects": json.dumps(objects),
            "request": json.dumps(request),
            "csrf_token": self.csrf_token
        }

        response = self.method("post", "runner/", headers, payload, raise_not_200=True)
        json_response = response.json()
        if not (resp := json_response.get("response")):
            raise exceptions.MessageNotDeliveredError(response, None, chat_id)

        if (error_text := resp.get("error")) is not None:
            raise exceptions.MessageNotDeliveredError(response, error_text, chat_id)

        mes = json_response["objects"][0]["data"]["messages"][-1]
        parser = BeautifulSoup(mes["html"], "html.parser")
        try:
            if image_link := parser.find("a", {"class": "chat-img-link"}):
                image_link = image_link.get("href")
                message_text = None
            else:
                message_text = parser.find("div", {"class": "chat-msg-text"}).text.replace(self.__bot_character, "", 1)
        except Exception as e:
            logger.debug("SEND_MESSAGE RESPONSE")
            logger.debug(response.content.decode())
            raise e

        message_obj = types.Message(int(mes["id"]), message_text, chat_id, chat_name, self.username, self.id,
                                    mes["html"], image_link)
        if self.runner and isinstance(chat_id, int):
            if add_to_ignore_list:
                self.runner.mark_as_by_bot(chat_id, message_obj.id)
            if update_last_saved_message:
                self.runner.update_last_message(chat_id, message_text)
        return message_obj

    def send_image(self, chat_id: int, image: int | str | IO[bytes], chat_name: Optional[str] = None,
                   add_to_ignore_list: bool = True, update_last_saved_message: bool = False) -> types.Message:
        """
        Отправляет изображение в чат. Доступно только для личных чатов.

        :param chat_id: ID чата.
        :type chat_id: :obj:`int`

        :param image: ID изображения / путь до изображения / изображение в виде байтов.
            Если передан путь до изображения или представление изображения в виде байтов, сначала оно будет выгружено
            с помощью метода :meth:`FunPayAPI.account.Account.upload_image`.
        :type image: :obj:`int` or :obj:`str` or :obj:`bytes`

        :param chat_name: Название чата (никнейм собеседника). Нужен для возвращаемого объекта.
        :type chat_name: :obj:`str` or :obj:`None`, опционально

        :param add_to_ignore_list: добавлять ли ID отправленного сообщения в игнорируемый список Runner'а?
        :type add_to_ignore_list: :obj:`bool`, опционально

        :param update_last_saved_message: обновлять ли последнее сохраненное сообщение на отправленное в Runner'е?
        :type update_last_saved_message: :obj:`bool`, опционально

        :return: объект отправленного сообщения.
        :rtype: :class:`FunPayAPI.types.Message`
        """
        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()

        if not isinstance(image, int):
            image = self.upload_image(image)
        result = self.send_message(chat_id, None, chat_name, image, add_to_ignore_list, update_last_saved_message)
        return result

    def send_review(self, order_id: str, text: str, rating: Literal[1, 2, 3, 4, 5] = 5) -> str:
        """
        Отправляет / редактирует отзыв / ответ на отзыв.

        :param order_id: ID заказа.
        :type order_id: :obj:`str`

        :param text: текст отзыва.
        :type text: :obj:`str`

        :param rating: рейтинг (от 1 до 5).
        :type rating: :obj:`int`, опционально

        :return: ответ FunPay (HTML-код блока отзыва).
        :rtype: :obj:`str`
        """
        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()

        headers = {
            "accept": "*/*",
            "x-requested-with": "XMLHttpRequest"
        }
        payload = {
            "authorId": self.id,
            "text": text,
            "rating": rating,
            "csrf_token": self.csrf_token,
            "orderId": order_id
        }

        response = self.method("post", "orders/review", headers, payload)
        if response.status_code == 400:
            json_response = response.json()
            msg = json_response.get("msg")
            raise exceptions.FeedbackEditingError(response, msg, order_id)
        elif response.status_code != 200:
            raise exceptions.RequestFailedError(response)

        return response.json().get("content")

    def delete_review(self, order_id: str) -> str:
        """
        Удаляет отзыв / ответ на отзыв.

        :param order_id: ID заказа.
        :type order_id: :obj:`str`

        :return: ответ FunPay (HTML-код блока отзыва).
        :rtype: :obj:`str`
        """
        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()

        headers = {
            "accept": "*/*",
            "x-requested-with": "XMLHttpRequest"
        }
        payload = {
            "authorId": self.id,
            "csrf_token": self.csrf_token,
            "orderId": order_id
        }

        response = self.method("post", "orders/reviewDelete", headers, payload)

        if response.status_code == 400:
            json_response = response.json()
            msg = json_response.get("msg")
            raise exceptions.FeedbackEditingError(response, msg, order_id)
        elif response.status_code != 200:
            raise exceptions.RequestFailedError(response)

        return response.json().get("content")

    def refund(self, order_id):
        """
        Оформляет возврат средств за заказ.

        :param order_id: ID заказа.
        :type order_id: :obj:`str`
        """
        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()

        headers = {
            "accept": "*/*",
            "content-type": "application/x-www-form-urlencoded; charset=UTF-8",
            "x-requested-with": "XMLHttpRequest",
        }

        payload = {
            "id": order_id,
            "csrf_token": self.csrf_token
        }

        response = self.method("post", "orders/refund", headers, payload, raise_not_200=True)

        if response.json().get("error"):
            raise exceptions.RefundError(response, response.json().get("msg"), order_id)

    def withdraw(self, currency: enums.Currency, wallet: enums.Wallet, amount: int | float, address: str) -> float:
        """
        Отправляет запрос на вывод средств.

        :param currency: валюта.
        :type currency: :class:`FunPayAPI.common.enums.Currency`

        :param wallet: тип кошелька.
        :type wallet: :class:`FunPayAPI.common.enums.Wallet`

        :param amount: кол-во средств.
        :type amount: :obj:`int` or :obj:`float`

        :param address: адрес кошелька.
        :type address: :obj:`str`

        :return: кол-во выведенных средств с учетом комиссии FunPay.
        :rtype: :obj:`float`
        """
        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()

        currencies = {
            enums.Currency.RUB: "rub",
            enums.Currency.USD: "usd",
            enums.Currency.EUR: "eur"
        }

        wallets = {
            enums.Wallet.QIWI: "qiwi",
            enums.Wallet.YOUMONEY: "fps",
            enums.Wallet.BINANCE: "binance",
            enums.Wallet.TRC: "usdt_trc",
            enums.Wallet.CARD_RUB: "card_rub",
            enums.Wallet.CARD_USD: "card_usd",
            enums.Wallet.CARD_EUR: "card_eur",
            enums.Wallet.WEBMONEY: "wmz"
        }
        headers = {
            "accept": "*/*",
            "x-requested-with": "XMLHttpRequest"
        }
        payload = {
            "csrf_token": self.csrf_token,
            "currency_id": currencies[currency],
            "ext_currency_id": wallets[wallet],
            "wallet": address,
            "amount_int": str(amount)
        }
        response = self.method("post", "withdraw/withdraw", headers, payload, raise_not_200=True)
        json_response = response.json()
        if json_response.get("error"):
            error_message = json_response.get("msg")
            raise exceptions.WithdrawError(response, error_message)
        return float(json_response.get("amount_ext"))

    def get_raise_modal(self, category_id: int) -> dict:
        """
        Отправляет запрос на получение modal-формы для поднятия лотов категории (игры).
        !ВНИМАНИЕ! Если на аккаунте только 1 подкатегория, относящаяся переданной категории (игре),
        то FunPay поднимет лоты данной подкатегории без отправления modal-формы с выбором других подкатегорий.

        :param category_id: ID категории (игры).
        :type category_id: :obj:`int`

        :return: ответ FunPay.
        :rtype: :obj:`dict`
        """
        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()
        category = self.get_category(category_id)
        subcategory = category.get_subcategories()[0]
        headers = {
            "accept": "*/*",
            "content-type": "application/x-www-form-urlencoded; charset=UTF-8",
            "x-requested-with": "XMLHttpRequest"
        }
        payload = {
            "game_id": category_id,
            "node_id": subcategory.id
        }
        response = self.method("post", "https://funpay.com/lots/raise", headers, payload, raise_not_200=True)
        json_response = response.json()
        return json_response

    def raise_lots(self, category_id: int, subcategories: Optional[list[int | types.SubCategory]] = None,
                   exclude: list[int] | None = None) -> bool:
        """
        Поднимает все лоты всех подкатегорий переданной категории (игры).

        :param category_id: ID категории (игры).
        :type category_id: :obj:`int`

        :param subcategories: список подкатегорий, которые необходимо поднять. Если не указаны, поднимутся все
            подкатегории переданной категории.
        :type subcategories: :obj:`list` of :obj:`int` or :class:`FunPayAPI.types.SubCategory`

        :param exclude: ID подкатегорий, которые не нужно поднимать.
        :type exclude: :obj:`list` of :obj:`int`, опционально.

        :return: `True`
        :rtype: :obj:`bool`
        """
        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()
        if not (category := self.get_category(category_id)):
            raise Exception("Not Found")  # todo

        exclude = exclude or []
        if subcategories:
            subcats = []
            for i in subcategories:
                if isinstance(i, types.SubCategory):
                    if i.type is types.SubCategoryTypes.COMMON and i.category.id == category.id and i.id not in exclude:
                        subcats.append(i)
                else:
                    if not (subcat := category.get_subcategory(types.SubCategoryTypes.COMMON, i)):
                        continue
                    subcats.append(subcat)
        else:
            subcats = [i for i in category.get_subcategories() if i.type is types.SubCategoryTypes.COMMON and i.id not in exclude]

        headers = {
            "accept": "*/*",
            "content-type": "application/x-www-form-urlencoded; charset=UTF-8",
            "x-requested-with": "XMLHttpRequest"
        }
        payload = {
            "game_id": category_id,
            "node_id": subcats[0].id,
            "node_ids[]": [i.id for i in subcats]
        }

        response = self.method("post", "lots/raise", headers, payload, raise_not_200=True)
        json_response = response.json()
        logger.debug(f"Ответ FunPay (поднятие категорий): {json_response}.")
        if not json_response.get("error"):
            return True
        elif json_response.get("error") and json_response.get("msg") and "Подождите" in json_response.get("msg"):
            wait_time = utils.parse_wait_time(json_response.get("msg"))
            raise exceptions.RaiseError(response, category, json_response.get("MSG"), wait_time)
        else:
            raise exceptions.RaiseError(response, category, None, None)

    def get_user(self, user_id: int) -> types.UserProfile:
        """
        Парсит страницу пользователя.

        :param user_id: ID пользователя.
        :type user_id: :obj:`int`

        :return: объект профиля пользователя.
        :rtype: :class:`FunPayAPI.types.UserProfile`
        """
        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()

        response = self.method("get", f"users/{user_id}/", {"accept": "*/*"}, {}, raise_not_200=True)
        html_response = response.content.decode()
        parser = BeautifulSoup(html_response, "html.parser")

        username = parser.find("div", {"class": "user-link-name"})
        if not username:
            raise exceptions.UnauthorizedError(response)

        username = parser.find("span", {"class": "mr4"}).text
        user_status = parser.find("span", {"class": "media-user-status"})
        user_status = user_status.text if user_status else ""
        avatar_link = parser.find("div", {"class": "avatar-photo"}).get("style").split("(")[1].split(")")[0]
        avatar_link = avatar_link if avatar_link.startswith("https") else f"https://funpay.com{avatar_link}"
        banned = bool(parser.find("span", {"class": "label label-danger"}))
        user_obj = types.UserProfile(user_id, username, avatar_link, "Онлайн" in user_status, banned, html_response)

        subcategories_divs = parser.find_all("div", {"class": "offer-list-title-container"})

        if not subcategories_divs:
            return user_obj

        for i in subcategories_divs:
            subcategory_link = i.find("h3").find("a").get("href")
            subcategory_id = int(subcategory_link.split("/")[-2])
            subcategory_type = types.SubCategoryTypes.CURRENCY if "chips" in subcategory_link else \
                types.SubCategoryTypes.COMMON
            subcategory_obj = self.get_subcategory(subcategory_type, subcategory_id)
            if not subcategory_obj:
                continue

            offers = i.parent.find_all("a", {"class": "tc-item"})
            for j in offers:
                offer_id = j["href"].split("id=")[1]
                description = j.find("div", {"class": "tc-desc-text"})
                description = description.text if description else None
                server = j.find("div", {"class": "tc-server hidden-xxs"})
                if not server:
                    server = j.find("div", {"class": "tc-server hidden-xs"})
                server = server.text if server else None

                if subcategory_obj.type is types.SubCategoryTypes.COMMON:
                    price = float(j.find("div", {"class": "tc-price"})["data-s"])
                else:
                    price = float(j.find("div", {"class": "tc-price"}).find("div").text.split(" ")[0])

                lot_obj = types.LotShortcut(offer_id, server, description, price, subcategory_obj, str(j))
                user_obj.add_lot(lot_obj)
        return user_obj

    def get_chat(self, chat_id: int) -> types.Chat:
        """
        Получает информацию о личном чате.

        :param chat_id: ID чата.
        :type chat_id: :obj:`int`

        :return: объект чата.
        :rtype: :class:`FunPayAPI.types.Chat`
        """
        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()

        response = self.method("get", f"chat/?node={chat_id}", {"accept": "*/*"}, {}, raise_not_200=True)
        html_response = response.content.decode()
        parser = BeautifulSoup(html_response, "html.parser")
        if (name := parser.find("div", {"class": "chat-header"}).find("div", {"class": "media-user-name"}).find("a").text) == "Чат":
            raise Exception("chat not found")  # todo

        if not (chat_panel := parser.find("div", {"class": "param-item chat-panel"})):
            text, link = None, None
        else:
            a = chat_panel.find("a")
            text, link = a.text, a["href"]

        history = self.get_chat_history(chat_id, interlocutor_username=name)
        return types.Chat(chat_id, name, link, text, html_response, history)

    def get_order(self, order_id: str) -> types.Order:
        """
        Получает полную информацию о заказе.

        :param order_id: ID заказа.
        :type order_id: :obj:`str`

        :return: объекст заказа.
        :rtype: :class:`FunPayAPI.types.Order`
        """
        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()
        headers = {
            "accept": "*/*"
        }
        response = self.method("get", f"orders/{order_id}/", headers, {}, raise_not_200=True)
        html_response = response.content.decode()
        parser = BeautifulSoup(html_response, "html.parser")
        username = parser.find("div", {"class": "user-link-name"})
        if not username:
            raise exceptions.UnauthorizedError(response)

        if (span := parser.find("span", {"class": "text-warning"})) and span.text == "Возврат":
            status = types.OrderStatuses.REFUNDED
        elif (span := parser.find("span", {"class": "text-success"})) and span.text == "Закрыт":
            status = types.OrderStatuses.CLOSED
        else:
            status = types.OrderStatuses.PAID

        short_description = None
        full_description = None
        sum_ = None
        subcategory = None
        for div in parser.find_all("div", {"class": "param-item"}):
            if not (h := div.find("h5")):
                continue
            if h.text == "Краткое описание":
                short_description = div.find("div").text
            elif h.text == "Подробное описание":
                full_description = div.find("div").text
            elif h.text == "Сумма":
                sum_ = float(div.find("span").text)
            elif h.text == "Категория":
                subcategory_link = div.find("a").get("href")
                subcategory_split = subcategory_link.split("/")
                subcategory_id = int(subcategory_split[-2])
                subcategory_type = types.SubCategoryTypes.COMMON if "lots" in subcategory_link else \
                    types.SubCategoryTypes.CURRENCY
                subcategory = self.get_subcategory(subcategory_type, subcategory_id)

        chat = parser.find("div", {"class": "chat-header"})
        chat_link = chat.find("div", {"class": "media-user-name"}).find("a")
        interlocutor_name = chat_link.text
        interlocutor_id = int(chat_link.get("href").split("/")[-2])
        nav_bar = parser.find("ul", {"class": "nav navbar-nav navbar-right logged"})
        active_item = nav_bar.find("li", {"class": "active"})
        if "Продажи" in active_item.find("a").text.strip():
            buyer_id, buyer_username = interlocutor_id, interlocutor_name
            seller_id, seller_username = self.id, self.username
        else:
            buyer_id, buyer_username = self.id, self.username
            seller_id, seller_username = interlocutor_id, interlocutor_name

        review_obj = parser.find("div", {"class": "order-review"})
        if not (stars_obj := review_obj.find("div", {"class": "rating"})):
            stars, text,  = None, None
        else:
            stars = int(stars_obj.find("div").get("class")[0].split("rating")[1])
            text = review_obj.find("div", {"class": "review-item-text"}).text.strip()

        if not (reply_obj := review_obj.find("div", {"class": "review-item-answer review-compiled-reply"})):
            reply = None
        else:
            reply = reply_obj.find("div").text.strip()

        if all([not text, not reply]):
            review = None
        else:
            review = types.Review(stars, text, reply, False, str(reply_obj), order_id, buyer_username, buyer_id)

        order = types.Order(order_id, status, subcategory, short_description, full_description, sum_,
                            buyer_id, buyer_username, seller_id, seller_username, html_response, review)
        return order

    def get_sells(self, start_from: str | None = None, include_paid: bool = True, include_closed: bool = True,
                  include_refunded: bool = True, exclude_ids: list[str] | None = None,
                  id: Optional[int] = None, buyer: Optional[str] = None,
                  state: Optional[Literal["closed", "paid", "refunded"]] = None, game: Optional[int] = None,
                  section: Optional[str] = None, server: Optional[int] = None,
                  side: Optional[int] = None, **more_filters) -> tuple[str | None, list[types.OrderShortcut]]:
        """
        Получает и парсит список заказов со страницы https://funpay.com/orders/trade

        :param start_from: ID заказа, с которого начать список (ID заказа должен быть без '#'!).
        :type start_from: :obj:`str`

        :param include_paid: включить ли в список заказы, ожидающие выполнения?
        :type include_paid: :obj:`bool`, опционально

        :param include_closed: включить ли в список закрытые заказы?
        :type include_closed: :obj:`bool`, опционально

        :param include_refunded: включить ли в список заказы, за которые запрошен возврат средств?
        :type include_refunded: :obj:`bool`, опционально

        :param exclude_ids: исключить заказы с ID из списка (ID заказа должен быть без '#'!).
        :type exclude_ids: :obj:`list` of :obj:`str`, опционально

        :param id: ID заказа.
        :type id: :obj:`int`, опционально

        :param buyer: никнейм покупателя.
        :type buyer: :obj:`str`, опционально

        :param state: статус заказа.
        :type: :obj:`str` `paid`, `closed` or `refunded`, опционально

        :param game: ID игры.
        :type game: :obj:`int`, опционально

        :param section: ID категории в формате `<тип лота>-<ID категории>`.\n
            Типы лотов:\n
            * `lot` - стандартный лот (например: `lot-256`)\n
            * `chip` - игровая валюта (например: `chip-4471`)\n
        :type section: :obj:`str`, опционально

        :param server: ID сервера.
        :type server: :obj:`int`, опционально

        :param side: ID стороны (платформы).
        :type side: :obj:`int`, опционально.

        :param more_filters: доп. фильтры.

        :return: (ID след. заказа (для start_from), список заказов)
        :rtype: :obj:`tuple` (:obj:`str` or :obj:`None`, :obj:`list` of :class:`FunPayAPI.types.OrderShortcut`)
        """
        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()

        exclude_ids = exclude_ids or []
        filters = {"id": id, "buyer": buyer, "state": state, "game": game, "section": section, "server": server,
                   "side": side}
        filters = {name: filters[name] for name in filters if filters[name]}
        filters.update(more_filters)

        link = "https://funpay.com/orders/trade?"
        for name in filters:
            link += f"{name}={filters[name]}&"
        link = link[:-1]

        if start_from:
            filters["continue"] = start_from

        response = self.method("post" if start_from else "get", link, {}, filters, raise_not_200=True)
        html_response = response.content.decode()

        parser = BeautifulSoup(html_response, "html.parser")
        check_user = parser.find("div", {"class": "content-account content-account-login"})
        if check_user:
            raise exceptions.UnauthorizedError(response)

        next_order_id = parser.find("input", {"type": "hidden", "name": "continue"})
        next_order_id = next_order_id.get("value") if next_order_id else None

        order_divs = parser.find_all("a", {"class": "tc-item"})
        if not order_divs:
            return None, []

        sells = []
        for div in order_divs:
            classname = div.get("class")
            if "warning" in classname:
                if not include_refunded:
                    continue
                order_status = types.OrderStatuses.REFUNDED
            elif "info" in classname:
                if not include_paid:
                    continue
                order_status = types.OrderStatuses.PAID
            else:
                if not include_closed:
                    continue
                order_status = types.OrderStatuses.CLOSED

            order_id = div.find("div", {"class": "tc-order"}).text[1:]
            if order_id in exclude_ids:
                continue

            description = div.find("div", {"class": "order-desc"}).find("div").text
            price = float(div.find("div", {"class": "tc-price"}).text.split(" ")[0])

            buyer_div = div.find("div", {"class": "media-user-name"}).find("span")
            buyer_username = buyer_div.text
            buyer_id = int(buyer_div.get("data-href")[:-1].split("https://funpay.com/users/")[1])
            subcategory_name = div.find("div", {"class": "text-muted"}).text

            now = datetime.now()
            order_date_text = div.find("div", {"class": "tc-date-time"}).text
            if "сегодня" in order_date_text:  # сегодня, ЧЧ:ММ
                h, m = order_date_text.split(", ")[1].split(":")
                order_date = datetime(now.year, now.month, now.day, int(h), int(m))
            elif "вчера" in order_date_text:  # вчера, ЧЧ:ММ
                h, m = order_date_text.split(", ")[1].split(":")
                temp = now - timedelta(days=1)
                order_date = datetime(temp.year, temp.month, temp.day, int(h), int(m))
            elif order_date_text.count(" ") == 2:  # ДД месяца, ЧЧ:ММ
                split = order_date_text.split(", ")
                day, month = split[0].split()
                day, month = int(day), utils.MONTHS[month]
                h, m = split[1].split(":")
                order_date = datetime(now.year, month, day, int(h), int(m))
            else:  # ДД месяца ГГГГ, ЧЧ:ММ
                split = order_date_text.split(", ")
                day, month, year = split[0].split()
                day, month, year = int(day), utils.MONTHS[month], int(year)
                h, m = split[1].split(":")
                order_date = datetime(year, month, day, int(h), int(m))

            order_obj = types.OrderShortcut(order_id, description, price, buyer_username, buyer_id, order_status,
                                            order_date, subcategory_name, str(div))
            sells.append(order_obj)

        return next_order_id, sells

    def add_chats(self, chats: list[types.ChatShortcut]):
        """
        Сохраняет чаты.

        :param chats: объекты чатов.
        :type chats: :obj:`list` of :class:`FunPayAPI.types.ChatShortcut`
        """
        for i in chats:
            self.__saved_chats[i.id] = i

    def request_chats(self) -> list[types.ChatShortcut]:
        """
        Запрашивает чаты и парсит их.

        :return: объекты чатов (не больше 50).
        :rtype: :obj:`list` of :class:`FunPayAPI.types.ChatShortcut`
        """
        chats = {
            "type": "chat_bookmarks",
            "id": self.id,
            "tag": utils.random_tag(),
            "data": False
        }
        payload = {
            "objects": json.dumps([chats]),
            "request": False,
            "csrf_token": self.csrf_token
        }
        headers = {
            "accept": "*/*",
            "content-type": "application/x-www-form-urlencoded; charset=UTF-8",
            "x-requested-with": "XMLHttpRequest"
        }
        response = self.method("post", "https://funpay.com/runner/", headers, payload, raise_not_200=True)
        json_response = response.json()

        msgs = ""
        for obj in json_response["objects"]:
            if obj.get("type") != "chat_bookmarks":
                continue
            msgs = obj["data"]["html"]
        if not msgs:
            return []

        parser = BeautifulSoup(msgs, "html.parser")
        chats = parser.find_all("a", {"class": "contact-item"})
        chats_objs = []

        for msg in chats:
            chat_id = int(msg["data-id"])
            last_msg_text = msg.find("div", {"class": "contact-item-message"}).text
            unread = True if "unread" in msg.get("class") else False
            chat_with = msg.find("div", {"class": "media-user-name"}).text
            chat_obj = types.ChatShortcut(chat_id, chat_with, last_msg_text, unread, str(msg))
            chats_objs.append(chat_obj)
        return chats_objs

    def get_chats(self, update: bool = False) -> dict[int, types.ChatShortcut]:
        """
        Возвращает словарь с сохраненными чатами ({id: types.ChatShortcut})

        :param update: обновлять ли предварительно список чатов с помощью доп. запроса?
        :type update: :obj:`bool`, опционально

        :return: словарь с сохраненными чатами.
        :rtype: :obj:`dict` {:obj:`int`: :class:`FunPayAPi.types.ChatShortcut`}
        """
        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()
        if update:
            chats = self.request_chats()
            self.add_chats(chats)
        return self.__saved_chats

    def get_chat_by_name(self, name: str, make_request: bool = False) -> types.ChatShortcut | None:
        """
        Возвращает чат по его названию (если он сохранен).

        :param name: название чата.
        :type name: :obj:`str`

        :param make_request: обновить ли сохраненные чаты, если чат не был найден?
        :type make_request: :obj:`bool`, опционально

        :return: объект чата или :obj:`None`, если чат не был найден.
        :rtype: :class:`FunPayAPI.types.ChatShortcut` or :obj:`None`
        """
        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()

        for i in self.__saved_chats:
            if self.__saved_chats[i].name == name:
                return self.__saved_chats[i]

        if make_request:
            self.add_chats(self.request_chats())
            return self.get_chat_by_name(name)
        else:
            return None

    def get_chat_by_id(self, chat_id: int, make_request: bool = False) -> types.ChatShortcut | None:
        """
        Возвращает личный чат по его ID (если он сохранен).

        :param chat_id: ID чата.
        :type chat_id: :obj:`int`

        :param make_request: обновить ли сохраненные чаты, если чат не был найден?
        :type make_request: :obj:`bool`, опционально

        :return: объект чата или :obj:`None`, если чат не был найден.
        :rtype: :class:`FunPayAPI.types.ChatShortcut` or :obj:`None`
        """
        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()

        if not make_request or chat_id in self.__saved_chats:
            return self.__saved_chats.get(chat_id)

        self.add_chats(self.request_chats())
        return self.get_chat_by_id(chat_id)

    def get_lot_fields(self, lot_id: int) -> types.LotFields:
        """
        Получает все поля лота.

        :param lot_id: ID лота.
        :type lot_id: :obj:`int`

        :return: объект с полями лота.
        :rtype: :class:`FunPayAPI.types.LotFields`
        """
        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()
        headers = {
            "accept": "*/*",
            "content-type": "application/json",
            "x-requested-with": "XMLHttpRequest",
        }
        response = self.method("get", f"lots/offerEdit?offer={lot_id}", headers, {}, raise_not_200=True)

        json_response = response.json()
        bs = BeautifulSoup(json_response["html"], "html.parser")

        result = {"active": "", "deactivate_after_sale": ""}
        result.update({field["name"]: field.get("value") or "" for field in bs.find_all("input")
                       if field["name"] not in ["active", "deactivate_after_sale"]})
        result.update({field["name"]: field.text or "" for field in bs.find_all("textarea")})
        result.update({field["name"]: field.find("option", selected=True)["value"] for field in bs.find_all("select")})
        result.update({field["name"]: "on" for field in bs.find_all("input", {"type": "checkbox"}, checked=True)})
        return types.LotFields(lot_id, result)

    def save_lot(self, lot_fields: types.LotFields):
        """
        Сохраняет лот на FunPay.

        :param lot_fields: объект с полями лота.
        :type lot_fields: :class:`FunPayAPI.types.LotFields`
        """
        if not self.is_initiated:
            raise exceptions.AccountNotInitiatedError()
        headers = {
            "accept": "*/*",
            "content-type": "application/x-www-form-urlencoded; charset=UTF-8",
            "x-requested-with": "XMLHttpRequest",
        }
        fields = lot_fields.renew_fields().fields
        fields["location"] = "trade"

        response = self.method("post", "lots/offerSave", headers, fields, raise_not_200=True)
        json_response = response.json()
        if json_response.get("error"):
            raise exceptions.LotSavingError(response, json_response.get("error"), lot_fields.lot_id)

    def get_category(self, category_id: int) -> types.Category | None:
        """
        Возвращает объект категории (игры).

        :param category_id: ID категории (игры).
        :type category_id: :obj:`int`

        :return: объект категории (игры) или :obj:`None`, если категория не была найдена.
        :rtype: :class:`FunPayAPI.types.Category` or :obj:`None`
        """
        return self.__sorted_categories.get(category_id)

    @property
    def categories(self) -> list[types.Category]:
        """
        Возвращает все категории (игры) FunPay (парсятся при первом выполнении метода :meth:`FunPayAPI.account.Account.get`).

        :return: все категории (игры) FunPay.
        :rtype: :obj:`list` of :class:`FunPayAPI.types.Category`
        """
        return self.__categories

    def get_sorted_categories(self) -> dict[int, types.Category]:
        """
        Возвращает все категории (игры) FunPay в виде словаря {ID: категория}
        (парсятся при первом выполнении метода :meth:`FunPayAPI.account.Account.get`).

        :return: все категории (игры) FunPay в виде словаря {ID: категория}
        :rtype: :obj:`dict` {:obj:`int`: :class:`FunPayAPI.types.Category`}
        """
        return self.__sorted_categories

    def get_subcategory(self, subcategory_type: types.SubCategoryTypes,
                        subcategory_id: int) -> types.SubCategory | None:
        """
        Возвращает объект подкатегории.

        :param subcategory_type: тип подкатегории.
        :type subcategory_type: :class:`FunPayAPI.common.enums.SubCategoryTypes`

        :param subcategory_id: ID подкатегории.
        :type subcategory_id: :obj:`int`

        :return: объект подкатегории или :obj:`None`, если подкатегория не была найдена.
        :rtype: :class:`FunPayAPI.types.SubCategory` or :obj:`None`
        """
        return self.__sorted_subcategories[subcategory_type].get(subcategory_id)

    @property
    def subcategories(self) -> list[types.SubCategory]:
        """
        Возвращает все подкатегории FunPay (парсятся при первом выполнении метода Account.get).

        :return: все подкатегории FunPay.
        :rtype: :obj:`list` of :class:`FunPayAPI.types.SubCategory`
        """
        return self.__subcategories

    def get_sorted_subcategories(self) -> dict[types.SubCategoryTypes, dict[int, types.SubCategory]]:
        """
        Возвращает все подкатегории FunPay в виде словаря {тип подкатегории: {ID: подкатегория}}
        (парсятся при первом выполнении метода Account.get).

        :return: все подкатегории FunPay в виде словаря {тип подкатегории: {ID: подкатегория}}
        :rtype: :obj:`dict` {:class:`FunPayAPI.common.enums.SubCategoryTypes`: :obj:`dict` {:obj:`int` :class:`FunPayAPI.types.SubCategory`}}
        """
        return self.__sorted_subcategories

    @property
    def is_initiated(self) -> bool:
        """
        Инициализирован ли класс :class:`FunPayAPI.account.Account` с помощью метода :meth:`FunPayAPI.account.Account.get`?

        :return: :obj:`True`, если да, :obj:`False`, если нет.
        :rtype: :obj:`bool`
        """
        return self.__initiated

    def __setup_categories(self, html: str):
        """
        Парсит категории и подкатегории с основной страницы и добавляет их в свойства класса.

        :param html: HTML страница.
        """
        parser = BeautifulSoup(html, "html.parser")
        games_table = parser.find_all("div", {"class": "promo-game-list"})
        if not games_table:
            return

        games_table = games_table[1] if len(games_table) > 1 else games_table[0]
        games_divs = games_table.find_all("div", {"class": "promo-game-item"})
        if not games_divs:
            return

        for i in games_divs:
            game_id = int(i.find("div", {"class": "game-title"}).get("data-id"))
            game_title = i.find("a").text
            game_obj = types.Category(game_id, game_title)

            subcategories_divs = i.find_all("li")
            for j in subcategories_divs:
                subcategory_name = j.find("a").text
                link = j.find("a").get("href")
                subcategory_type = types.SubCategoryTypes.CURRENCY if "chips" in link else types.SubCategoryTypes.COMMON
                subcategory_id = int(link.split("/")[-2])

                subcategory_obj = types.SubCategory(subcategory_id, subcategory_name, subcategory_type, game_obj)
                game_obj.add_subcategory(subcategory_obj)
                self.__subcategories.append(subcategory_obj)
                self.__sorted_subcategories[subcategory_type][subcategory_id] = subcategory_obj

            self.__categories.append(game_obj)
            self.__sorted_categories[game_id] = game_obj

    def __parse_messages(self, json_messages: dict, chat_id: int | str,
                         interlocutor_id: Optional[int] = None, interlocutor_username: Optional[str] = None,
                         from_id: int = 0) -> list[types.Message]:
        messages = []
        ids = {self.id: self.username, 0: "FunPay"}
        badges = {}
        if interlocutor_id is not None:
            ids[interlocutor_id] = interlocutor_username

        for i in json_messages:
            if i["id"] < from_id:
                continue
            author_id = i["author"]
            parser = BeautifulSoup(i["html"], "html.parser")

            # Если ник или бейдж написавшего неизвестен, но есть блок с данными об авторе сообщения
            if None in [ids.get(author_id), badges.get(author_id)] and (author_div := parser.find("div", {"class": "media-user-name"})):
                if badges.get(author_id) is None:
                    badge = author_div.find("span")
                    badges[author_id] = badge.text if badge else 0
                if ids.get(author_id) is None:
                    author = author_div.find("a").text.strip()
                    ids[author_id] = author
                    if self.chat_id_private(chat_id) and author_id == interlocutor_id and not interlocutor_username:
                        interlocutor_username = author
                        ids[interlocutor_id] = interlocutor_username

            if self.chat_id_private and (image_link := parser.find("a", {"class": "chat-img-link"})):
                image_link = image_link.get("href")
                message_text = None
            else:
                image_link = None
                if author_id == 0:
                    message_text = parser.find("div", {"class": "alert alert-with-icon alert-info"}).text.strip()
                else:
                    message_text = parser.find("div", {"class": "chat-msg-text"}).text

            by_bot = False
            if not image_link and message_text.startswith(self.__bot_character):
                message_text = message_text.replace(self.__bot_character, "", 1)
                by_bot = True

            message_obj = types.Message(i["id"], message_text, chat_id, interlocutor_username,
                                        None, author_id, i["html"], image_link, determine_msg_type=False)
            message_obj.by_bot = by_bot
            message_obj.type = types.MessageTypes.NON_SYSTEM if author_id != 0 else message_obj.get_message_type()
            messages.append(message_obj)

        for i in messages:
            i.author = ids.get(i.author_id)
            i.chat_name = interlocutor_username
            i.badge = badges.get(i.author_id) if badges.get(i.author_id) != 0 else None
        return messages

    @staticmethod
    def chat_id_private(chat_id: int | str):
        return isinstance(chat_id, int) or PRIVATE_CHAT_ID_RE.fullmatch(chat_id)

    @property
    def bot_character(self) -> str:
        return self.__bot_character



from setuptools import setup, find_packages


setup(
   name='funpay-cfn',  # Вставьте уникальное имя вашей библиотеки
   version='0.1.5',
   author='cfn',
   author_email='cfn@cfn.com',
   description='cfn',
   long_description=open('README.md').read(),
   long_description_content_type='text/markdown',
   packages=find_packages(where="."),
   install_requires=[
       'requests_toolbelt',
       'beautifulsoup4',
       'requests',
   ],
   classifiers=[
       'Programming Language :: Python :: 3',
       'License :: OSI Approved :: MIT License',
       'Operating System :: OS Independent',
   ],
   python_requires='>=3.6',
)


from __future__ import annotations
import time
from ..common import utils
from ..common.enums import *
from .. import types


class BaseEvent:
    """
    Базовый класс события.

    :param runner_tag: тег Runner'а.
    :type runner_tag: :obj:`str`

    :param event_type: тип события.
    :type event_type: :class:`FunPayAPI.common.enums.EventTypes`

    :param event_time: время события (лучше не указывать, будет генерироваться автоматически).
    :type event_time: :obj:`int` or :obj:`float` or :obj:`None`, опционально.
    """
    def __init__(self, runner_tag: str, event_type: EventTypes, event_time: int | float | None = None):
        self.runner_tag = runner_tag
        self.type = event_type
        self.time = event_time if event_type is not None else time.time()


class InitialChatEvent(BaseEvent):
    """
    Класс события: обнаружен чат при первом запросе Runner'а.

    :param runner_tag: тег Runner'а.
    :type runner_tag: :obj:`str`

    :param chat_obj: объект обнаруженного чата.
    :type chat_obj: :class:`FunPayAPI.types.ChatShortcut`
    """
    def __init__(self, runner_tag: str, chat_obj: types.ChatShortcut):
        super(InitialChatEvent, self).__init__(runner_tag, EventTypes.INITIAL_CHAT)
        self.chat: types.ChatShortcut = chat_obj
        """Объект обнаруженного чата."""


class ChatsListChangedEvent(BaseEvent):
    """
    Класс события: список чатов и / или содержимое одного / нескольких чатов изменилось.

    :param runner_tag: тег Runner'а.
    :type runner_tag: :obj:`str`
    """
    def __init__(self, runner_tag: str):
        super(ChatsListChangedEvent, self).__init__(runner_tag, EventTypes.CHATS_LIST_CHANGED)
        # todo: добавить список всех чатов.


class LastChatMessageChangedEvent(BaseEvent):
    """
    Класс события: последнее сообщение в чате изменилось.

    :param runner_tag: тег Runner'а.
    :type runner_tag: :obj:`str`

    :param chat_obj: объект чата, в котором изменилось полседнее сообщение.
    :type chat_obj: :class:`FunPayAPI.types.ChatShortcut`
    """
    def __init__(self, runner_tag: str, chat_obj: types.ChatShortcut):
        super(LastChatMessageChangedEvent, self).__init__(runner_tag, EventTypes.LAST_CHAT_MESSAGE_CHANGED)
        self.chat: types.ChatShortcut = chat_obj
        """Объект чата, в котором изменилось полседнее сообщение."""


class NewMessageEvent(BaseEvent):
    """
    Класс события: в истории чата обнаружено новое сообщение.

    :param runner_tag: тег Runner'а.
    :type runner_tag: :obj:`str`

    :param message_obj: объект нового сообщения.
    :type message_obj: :class:`FunPayAPI.types.Message`

    :param stack: объект стэка событий новых собщений.
    :type stack: :class:`FunPayAPI.updater.events.MessageEventsStack` or :obj:`None`, опционально
    """
    def __init__(self, runner_tag: str, message_obj: types.Message, stack: MessageEventsStack | None = None):
        super(NewMessageEvent, self).__init__(runner_tag, EventTypes.NEW_MESSAGE)
        self.message: types.Message = message_obj
        """Объект нового сообщения."""
        self.stack: MessageEventsStack = stack
        """Объект стэка событий новых сообщений."""


class MessageEventsStack:
    """
    Данный класс представляет стэк событий новых сообщений.
    Нужен для того, чтобы сразу предоставить доступ ко всем событиям новых сообщений от одного пользователя и одного запроса Runner'а.
    """
    def __init__(self):
        self.__id = utils.random_tag()
        self.__stack = []

    def add_events(self, messages: list[NewMessageEvent]):
        """
        Добавляет события новых сообщений в стэк.

        :param messages: список событий новых сообщений.
        :type messages: :obj:`list` of :class:`FunPayAPI.updater.events.NewMessageEvent`
        """
        self.__stack.extend(messages)

    def get_stack(self) -> list[NewMessageEvent]:
        """
        Возвращает стэк событий новых сообщений.

        :return: стэк событий новых сообщений.
        :rtype: :obj:`list` of :class:`FunPayAPI.updater.events.NewMessageEvent`
        """
        return self.__stack

    def id(self) -> str:
        """
        Возвращает ID стэка (ID стега генерируется случайным образом при создании объекта).

        :return: ID стэка.
        :rtype: :obj:`str`
        """
        return self.__id


class InitialOrderEvent(BaseEvent):
    """
    Класс события: обнаружен заказ при первом запросе Runner'а.

    :param runner_tag: тег Runner'а.
    :type runner_tag: :obj:`str`

    :param order_obj: объект обнаруженного заказа.
    :type order_obj: :class:`FunPayAPI.types.OrderShortcut`
    """
    def __init__(self, runner_tag: str, order_obj: types.OrderShortcut):
        super(InitialOrderEvent, self).__init__(runner_tag, EventTypes.INITIAL_ORDER)
        self.order: types.OrderShortcut = order_obj
        """Объект обнаруженного заказа."""


class OrdersListChangedEvent(BaseEvent):
    """
    Класс события: список заказов и/или статус одного/нескольких заказов изменился.

    :param runner_tag: тег Runner'а.
    :type runner_tag: :obj:`str`

    :param purchases: кол-во незавершенных покупок.
    :type purchases: :obj:`int`

    :param sales: кол-во незавершенных продаж.
    :type sales: :obj:`int`
    """
    def __init__(self, runner_tag: str, purchases: int, sales: int):
        super(OrdersListChangedEvent, self).__init__(runner_tag, EventTypes.ORDERS_LIST_CHANGED)
        self.purchases: int = purchases
        """Кол-во незавершенных покупок."""
        self.sales: int = sales
        """Кол-во незавершенных продаж."""


class NewOrderEvent(BaseEvent):
    """
    Класс события: в списке заказов обнаружен новый заказ.

    :param runner_tag: тег Runner'а.
    :type runner_tag: :obj:`str`

    :param order_obj: объект нового заказа.
    :type order_obj: :class:`FunPayAPI.types.OrderShortcut`
    """
    def __init__(self, runner_tag: str, order_obj: types.OrderShortcut):
        super(NewOrderEvent, self).__init__(runner_tag, EventTypes.NEW_ORDER)
        self.order: types.OrderShortcut = order_obj
        """Объект нового заказа."""


class OrderStatusChangedEvent(BaseEvent):
    """
    Класс события: статус заказа изменился.

    :param runner_tag: тег Runner'а.
    :type runner_tag: :obj:`str`

    :param order_obj: объект измененного заказа.
    :type order_obj: :class:`FunPayAPI.types.OrderShortcut`
    """
    def __init__(self, runner_tag: str, order_obj: types.OrderShortcut):
        super(OrderStatusChangedEvent, self).__init__(runner_tag, EventTypes.ORDER_STATUS_CHANGED)
        self.order: types.OrderShortcut = order_obj
        """Объект измененного заказа."""
